from __future__ import annotations

from logging import getLogger
from typing import TYPE_CHECKING

from .base import HConfigBase
from .child import HConfigChild
from .models import Dump, DumpLine, ReferenceLocation

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator

    from hier_config.platforms.driver_base import HConfigDriverBase

logger = getLogger(__name__)

# Refactoring ideas:
# - Cases of children.index() could be replaced with an identity based approach.


class HConfig(HConfigBase):  # noqa: PLR0904
    """A class for representing and comparing Cisco like configurations in a
    hierarchical tree data structure.
    """

    __slots__ = ("_driver",)

    def __init__(self, driver: HConfigDriverBase) -> None:
        super().__init__()
        self._driver = driver

    def __str__(self) -> str:
        return "\n".join(str(c) for c in sorted(self.children))

    def __repr__(self) -> str:
        return f"HConfig(driver={self.driver.__class__.__name__}, lines={self.dump_simple()})"

    def __hash__(self) -> int:
        return hash(*self.children)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, HConfig):
            return NotImplemented

        return self.children == other.children

    @property
    def driver(self) -> HConfigDriverBase:
        return self._driver

    @property
    def real_indent_level(self) -> int:
        return -1

    @property
    def parent(self) -> HConfig:
        return self

    @property
    def root(self) -> HConfig:
        """Returns the HConfig object at the base of the tree."""
        return self

    @property
    def is_leaf(self) -> bool:
        """Returns True if there are no children and is not an instance of HConfig."""
        return False

    @property
    def is_branch(self) -> bool:
        """Returns True if there are children or is an instance of HConfig."""
        return True

    def instantiate_child(self, text: str) -> HConfigChild:
        return HConfigChild(self, text)

    @property
    def tags(self) -> frozenset[str]:
        """Recursive access to tags on all leaf nodes."""
        found_tags: set[str] = set()
        for child in self.children:
            found_tags.update(child.tags)
        return frozenset(found_tags)

    @tags.setter
    def tags(self, value: frozenset[str]) -> None:
        """Recursive access to tags on all leaf nodes."""
        for child in self.children:
            child.tags = value

    def merge(self, other: HConfig | Iterable[HConfig]) -> HConfig:
        """Merges other HConfig objects into this one."""
        other_configs = (other,) if isinstance(other, HConfig) else other

        for other_config in other_configs:
            for child in other_config.children:
                self.add_deep_copy_of(child, merged=True)

        return self

    def add_children_deep(self, lines: Iterable[str]) -> HConfigChild:
        """Add child instances of HConfigChild deeply."""
        base: HConfig | HConfigChild = self
        for line in lines:
            base = base.add_child(line, return_if_present=True)
        if isinstance(base, HConfig):
            message = "base was an HConfig object for some reason."
            raise TypeError(message)
        return base

    def lineage(self) -> Iterator[HConfigChild]:  # noqa: PLR6301
        """Yields the lineage of parent objects, up to but excluding the root."""
        yield from ()

    def lines(self, *, sectional_exiting: bool = False) -> Iterable[str]:
        for child in sorted(self.children):
            yield from child.lines(sectional_exiting=sectional_exiting)

    def dump_simple(self, *, sectional_exiting: bool = False) -> tuple[str, ...]:
        return tuple(self.lines(sectional_exiting=sectional_exiting))

    def dump(self) -> Dump:
        """Dump loaded HConfig data."""
        return Dump(
            lines=tuple(
                DumpLine(
                    depth=c.depth(),
                    text=c.text,
                    tags=frozenset(c.tags),
                    comments=frozenset(c.comments),
                    new_in_config=c.new_in_config,
                )
                for c in self.all_children_sorted()
            ),
        )

    def depth(self) -> int:  # noqa: PLR6301
        """Returns the distance to the root HConfig object i.e. indent level."""
        return 0

    def difference(self, target: HConfig) -> HConfig:
        """Creates a new HConfig object with the config from self that is not in target."""
        return self._difference(target, HConfig(self.driver))

    def config_to_get_to(
        self,
        target: HConfig,
        delta: HConfig | None = None,
    ) -> HConfig:
        """Figures out what commands need to be executed to transition from self to target.
        self is the source data structure(i.e. the running_config),
        target is the destination(i.e. generated_config).
        """
        if delta is None:
            delta = HConfig(self.driver)

        return self._config_to_get_to(target, delta)

    def add_ancestor_copy_of(
        self,
        parent_to_add: HConfigChild,
    ) -> HConfig | HConfigChild:
        """Add a copy of the ancestry of parent_to_add to self
        and return the deepest child which is equivalent to parent_to_add.
        """
        base: HConfig | HConfigChild = self
        for parent in parent_to_add.lineage():
            base = base.add_shallow_copy_of(parent)

        return base

    def set_order_weight(self) -> HConfig:
        """Sets self.order integer on all children."""
        for child in self.all_children():
            for rule in self.driver.rules.ordering:
                if child.is_lineage_match(rule.match_rules):
                    child.order_weight = rule.weight
        return self

    def future(self, config: HConfig) -> HConfig:
        """EXPERIMENTAL - predict the future config after config is applied to self.

        The quality of this method's output will in part depend on how well
        the OS options are tuned. Ensuring that idempotency rules are accurate is
        especially important.
        """
        future_config = HConfig(self.driver)
        self._future(config, future_config)
        return future_config

    def with_tags(self, tags: Iterable[str]) -> HConfig:
        """Returns a new instance recursively containing children that only have a subset of tags."""
        return self._with_tags(frozenset(tags), HConfig(self.driver))

    def all_children_sorted_by_tags(
        self,
        include_tags: Iterable[str],
        exclude_tags: Iterable[str],
    ) -> Iterator[HConfigChild]:
        """Yield all children recursively that match include/exclude tags."""
        for child in sorted(self.children):
            yield from child.all_children_sorted_by_tags(include_tags, exclude_tags)

    def deep_copy(self) -> HConfig:
        """Return a copy of this object."""
        new_instance = HConfig(self.driver)
        for child in self.children:
            new_instance.add_deep_copy_of(child)
        return new_instance

    def unused_objects(self) -> Iterator[HConfigChild]:
        """Yield top-level children that are defined objects with no references.

        Uses ``self.driver.rules.unused_objects`` to identify object definitions,
        extract their names, and search for references across the config tree.
        Objects with zero references are yielded.
        """
        from re import search as _re_search  # noqa: PLC0415

        for rule in self.driver.rules.unused_objects:
            seen_names: set[str] = set()
            for definition in self.get_children_deep(rule.match_rules):
                match = _re_search(rule.name_re, definition.text)
                if not match:
                    continue
                name = match.group("name")
                if name in seen_names:
                    continue
                seen_names.add(name)

                if not self._is_object_referenced(name, rule.reference_locations):
                    yield definition

    def _is_object_referenced(
        self,
        name: str,
        reference_locations: tuple[ReferenceLocation, ...],
    ) -> bool:
        """Return True if *name* is found in any reference location."""
        from re import escape as _re_escape  # noqa: PLC0415
        from re import search as _re_search  # noqa: PLC0415

        for ref_location in reference_locations:
            pattern = ref_location.reference_re.format(name=_re_escape(name))
            for section in self.get_children_deep(ref_location.match_rules):
                if any(_re_search(pattern, d.text) for d in section.all_children()):
                    return True
        return False

    def _is_duplicate_child_allowed(self) -> bool:  # noqa: PLR6301
        """Determine if duplicate(identical text) children are allowed under the parent."""
        return False



from collections.abc import Iterable
from ipaddress import AddressValueError, IPv4Address, IPv4Interface
from re import sub

from hier_config.child import HConfigChild
from hier_config.platforms.functions import expand_range
from hier_config.platforms.models import (
    InterfaceDot1qMode,
    InterfaceDuplex,
    NACHostMode,
    StackMember,
    Vlan,
)
from hier_config.platforms.view_base import (
    ConfigViewInterfaceBase,
    HConfigViewBase,
)


class ConfigViewInterfaceCiscoIOS(ConfigViewInterfaceBase):  # noqa: PLR0904
    """Interface config view for Cisco IOS / IOS-XE."""

    @property
    def bundle_id(self) -> str | None:
        if channel_group := self.config.get_child(startswith="channel-group"):
            return channel_group.text.split()[1]
        return None

    @property
    def bundle_member_interfaces(self) -> Iterable[str]:
        raise NotImplementedError

    @property
    def bundle_name(self) -> str | None:
        if self.bundle_id:
            return f"{self._bundle_prefix}{self.bundle_id}"
        return None

    @property
    def description(self) -> str:
        if child := self.config.get_child(startswith="description "):
            return child.text.split(maxsplit=1)[1]
        return ""

    @property
    def duplex(self) -> InterfaceDuplex:
        if duplex := self.config.get_child(startswith="duplex "):
            return InterfaceDuplex(duplex.text.split()[1])
        return InterfaceDuplex.AUTO

    @property
    def enabled(self) -> bool:
        return not self.config.get_child(equals="shutdown")

    @property
    def has_nac(self) -> bool:
        return any(
            line in self.config.children
            for line in (
                "authentication port-control auto",
                "mab",
            )
        )

    @property
    def ipv4_interface(self) -> IPv4Interface | None:
        return next(iter(self.ipv4_interfaces), None)

    @property
    def ipv4_interfaces(self) -> Iterable[IPv4Interface]:
        for ipv4_address_obj in self.config.get_children(startswith="ip address "):
            ipv4_address = ipv4_address_obj.text.split()
            try:
                yield IPv4Interface("/".join(ipv4_address[2:4]))
            except AddressValueError:
                continue

    @property
    def is_bundle(self) -> bool:
        return self.name.lower().startswith(self._bundle_prefix)

    @property
    def is_loopback(self) -> bool:
        return self.name.lower().startswith("loopback")

    @property
    def is_subinterface(self) -> bool:
        return "." in self.name

    @property
    def is_svi(self) -> bool:
        return self.name.lower().startswith("vlan")

    @property
    def module_number(self) -> int | None:
        words = self.number.split("/", 1)
        if len(words) == 1:
            return None
        return int(words[0])

    @property
    def nac_control_direction_in(self) -> bool:
        """Determine if the interface has NAC control direction in configured."""
        return "authentication control-direction in" in self.config.children

    @property
    def nac_host_mode(self) -> NACHostMode | None:
        """Determine the NAC host mode."""
        mode = self.config.get_child(startswith="authentication host-mode ")
        if mode is None:
            return None

        mode_word = mode.text.split()[2]
        if mode_word == "multi-auth":
            return NACHostMode.MULTI_AUTH
        if mode_word == "multi-domain":
            return NACHostMode.MULTI_DOMAIN
        if mode_word == "multi-host":
            return NACHostMode.MULTI_HOST
        if mode_word == "single-host":
            return NACHostMode.SINGLE_HOST

        message = f"Unhandled NAC host mode: {mode_word}"
        raise ValueError(message)

    @property
    def nac_mab_first(self) -> bool:
        """Determine if the interface has NAC configured for MAB first."""
        return bool(self.config.get_child(equals="authentication order mab dot1x"))

    @property
    def nac_max_dot1x_clients(self) -> int:
        """Determine the max mab clients."""
        raise NotImplementedError

    @property
    def nac_max_mab_clients(self) -> int:
        """Determine the max mab clients."""
        raise NotImplementedError

    @property
    def name(self) -> str:
        return self.config.text.split()[1]

    @property
    def native_vlan(self) -> int | None:
        # It's configured as a sub-interface
        if self.is_subinterface and (
            vlan := self.config.get_child(startswith="encapsulation dot1Q ")
        ):
            return int(vlan.text.split()[2])

        # It's not a switchport
        if (
            self.config.get_child(equals="no switchport")
            or self.config.get_child(startswith="ip address ")
            or self.is_loopback
            or self.is_svi
        ):
            return None

        # It's configured as a trunk
        if self.config.get_child(equals="switchport mode trunk"):
            if vlan := self.config.get_child(
                startswith="switchport trunk native vlan ",
            ):
                return int(vlan.text.split()[4])

            return None

        # It's either dynamic or configured as an access port
        if vlan := self.config.get_child(startswith="switchport access vlan "):
            return int(vlan.text.split()[3])

        # Default VLAN
        return 1

    @property
    def number(self) -> str:
        return sub(r"^[a-zA-Z-]+", "", self.name)

    @property
    def parent_name(self) -> str | None:
        if self.is_subinterface:
            return self.name.split(".")[0]
        return None

    @property
    def poe(self) -> bool:
        return not self.config.get_child(equals="power inline never")

    @property
    def port_number(self) -> int:
        return int(self.name.split("/")[-1].split(".")[0])

    @property
    def speed(self) -> tuple[int, ...] | None:
        if speed := self.config.get_child(startswith="speed "):
            if speed.text == "auto":
                return None
            return (int(speed.text.split()[1]),)
        return None

    @property
    def subinterface_number(self) -> int | None:
        return int(self.name.split(".")[0 - 1]) if self.is_subinterface else None

    @property
    def tagged_all(self) -> bool:
        return bool(
            self.config.get_child(equals="switchport mode trunk")
            and not self.tagged_vlans,
        )

    @property
    def tagged_vlans(self) -> tuple[int, ...]:
        if child := self.config.get_child(
            re_search="^switchport trunk allowed vlan [0-9,-]+$",
        ):
            return expand_range(child.text.split()[4])
        return ()

    @property
    def vrf(self) -> str:
        if vrf := self.config.get_child(startswith="ip vrf forwarding "):
            return vrf.text.split()[3]
        return ""

    @property
    def _bundle_prefix(self) -> str:
        return "Port-channel"


class HConfigViewCiscoIOS(HConfigViewBase):
    """Full-tree config view for Cisco IOS / IOS-XE."""

    def dot1q_mode_from_vlans(
        self,
        untagged_vlan: int | None = None,
        tagged_vlans: tuple[int, ...] = (),
        *,
        tagged_all: bool = False,
    ) -> InterfaceDot1qMode | None:
        raise NotImplementedError

    @property
    def hostname(self) -> str | None:
        if child := self.config.get_child(startswith="hostname "):
            return child.text.split()[1].lower()
        return None

    @property
    def interface_names_mentioned(self) -> frozenset[str]:
        """Returns a set with all the interface names mentioned in the config."""
        return frozenset(model.name for model in self.interface_views)

    @property
    def interface_views(self) -> Iterable[ConfigViewInterfaceCiscoIOS]:
        for interface in self.interfaces:
            yield ConfigViewInterfaceCiscoIOS(interface)

    @property
    def interfaces(self) -> Iterable[HConfigChild]:
        return self.config.get_children(startswith="interface ")

    @property
    def ipv4_default_gw(self) -> IPv4Address | None:
        if gateway := self.config.get_child(startswith="ip default-gateway "):
            return IPv4Address(gateway.text.split()[2])
        return None

    @property
    def location(self) -> str:
        if location := self.config.get_child(startswith="snmp-server location "):
            return location.text.split(maxsplit=2)[2].replace('"', "")
        return ""

    @property
    def stack_members(self) -> Iterable[StackMember]:
        """Stacking
        member 1 type "JL123" mac-address abc123-abc123
        member 1 priority 255
        member 2 type "JL123" mac-address abc123-abc123
        member 2 priority 254
        ...
        """
        for member in self.config.get_children(re_search="^switch .* provision .*"):
            words = member.text.split()
            member_id = int(words[1])
            yield StackMember(
                id=member_id,
                priority=256 - member_id,
                mac_address=None,
                model=words[3],
            )

    @property
    def vlans(self) -> Iterable[Vlan]:
        yielded_vlans: set[int] = set()

        # Yield explicitly defined VLANs
        for child in self.config.get_children(re_search="^vlan [0-9,-]+$"):
            vlan_name = None
            if name := child.get_child(startswith="name "):
                _, vlan_name = name.text.split(maxsplit=1)
                vlan_name = vlan_name.replace('"', "")
            for vlan_id in expand_range(child.text.split()[1]):
                yielded_vlans.add(vlan_id)
                yield Vlan(
                    id=vlan_id,
                    name=vlan_name or None,
                )

        # Yield any remaining unnamed VLANs mentioned on interfaces
        for interface_view in self.interface_views:
            if (
                native_vlan := interface_view.native_vlan
            ) and native_vlan not in yielded_vlans:
                yielded_vlans.add(native_vlan)
                yield Vlan(
                    id=native_vlan,
                    name=None,
                )



from hier_config.child import HConfigChild
from hier_config.platforms.driver_base import HConfigDriverBase, HConfigDriverRules
from hier_config.platforms.functions import convert_to_set_commands


class HConfigDriverVYOS(HConfigDriverBase):  # pylint: disable=too-many-instance-attributes
    """Driver for VyOS (and compatible VyOS-based routers).

    Like the JunOS driver, converts hierarchical VyOS configuration into flat
    ``set``/``delete`` command syntax via a preprocessor.  Overrides
    ``declaration_prefix`` to ``"set "`` and ``negation_prefix`` to
    ``"delete "``.  Platform enum: ``Platform.VYOS``.
    """

    def swap_negation(self, child: HConfigChild) -> HConfigChild:
        """Swap negation of a `self.text`."""
        if child.text.startswith(self.negation_prefix):
            child.text = f"{self.declaration_prefix}{child.text_without_negation}"
        elif child.text.startswith(self.declaration_prefix):
            child.text = f"{self.negation_prefix}{child.text.removeprefix(self.declaration_prefix)}"

        return child

    @property
    def declaration_prefix(self) -> str:
        return "set "

    @property
    def negation_prefix(self) -> str:
        return "delete "

    @staticmethod
    def config_preprocessor(config_text: str) -> str:
        return convert_to_set_commands(config_text)

    @staticmethod
    def _instantiate_rules() -> HConfigDriverRules:
        return HConfigDriverRules()



def hp_procurve_expand_range(interface_range_str: str) -> tuple[str, ...]:
    """Expand interface ranges like 1/2-5,2/22-45."""
    interfaces: list[str] = []
    for interface_range in interface_range_str.split(","):
        _hp_procurve_expand_range_segment(interface_range, interfaces)
    if len(frozenset(interfaces)) != len(interfaces):
        message = (
            "the length of frozenset(interfaces) was not the same as len(interfaces)"
        )
        raise ValueError(message)
    return tuple(interfaces)


def _hp_procurve_expand_range_segment(
    interface_range: str,
    interfaces: list[str],
) -> None:
    start_stop = interface_range.split("-")
    if len(start_stop) != 2:
        interfaces.append(start_stop[0])
        return

    start_port_prefix = ""
    trk = "Trk"
    if start_stop[0].startswith(trk):
        stack_member = trk
        start_port_number = start_stop[0].removeprefix(trk)
        end_port_number = start_stop[1].removeprefix(trk)
    elif "/" in start_stop[0]:
        stack_member, start_port_number = start_stop[0].split("/")
        stack_member += "/"
        end_port_number = start_stop[1].split("/")[-1]
        # account for `interface 5/A1`
        for letter in ("A", "B", "C", "D"):
            if start_port_number.startswith(letter):
                start_port_prefix = letter
                start_port_number = start_port_number.removeprefix(letter)
                if not end_port_number.startswith(letter):
                    message = f"{letter=}, the end_port_number should start with the same letter"
                    raise ValueError(message)
                end_port_number = end_port_number.removeprefix(letter)
                break
    else:
        stack_member = ""
        start_port_number = start_stop[0]
        end_port_number = start_stop[1]

    interfaces.extend(
        f"{stack_member}{start_port_prefix}{port}"
        for port in range(int(start_port_number), int(end_port_number) + 1)
    )



from hier_config.models import (
    IdempotentCommandsAvoidRule,
    IdempotentCommandsRule,
    MatchRule,
    NegationDefaultWhenRule,
    NegationDefaultWithRule,
    PerLineSubRule,
)
from hier_config.platforms.driver_base import HConfigDriverBase, HConfigDriverRules


class HConfigDriverCiscoNXOS(HConfigDriverBase):
    """Driver for Cisco NX-OS (Nexus Operating System).

    Handles NX-OS quirks such as double-space in ``snmp-server location``,
    boot-image stripping, HSRP group-level idempotency, TCAM region
    assignments, and ``secondary`` IP address avoidance in idempotency checks.
    Platform enum: ``Platform.CISCO_NXOS``.
    """

    @staticmethod
    def _instantiate_rules() -> HConfigDriverRules:
        return HConfigDriverRules(
            per_line_sub=[
                PerLineSubRule(search="^Building configuration.*", replace=""),
                PerLineSubRule(search="^Current configuration.*", replace=""),
                PerLineSubRule(search="^ntp clock-period .*", replace=""),
                PerLineSubRule(
                    search="^snmp-server location  ",
                    replace="snmp-server location ",
                ),
                PerLineSubRule(search="^version.*", replace=""),
                PerLineSubRule(search="^boot (system|kickstart) .*", replace=""),
                PerLineSubRule(search="!.*", replace=""),
            ],
            idempotent_commands_avoid=[
                IdempotentCommandsAvoidRule(
                    match_rules=(
                        MatchRule(startswith="interface"),
                        MatchRule(re_search="ip address.*secondary"),
                    ),
                ),
            ],
            idempotent_commands=[
                IdempotentCommandsRule(
                    match_rules=(MatchRule(startswith="power redundancy-mode"),),
                ),
                IdempotentCommandsRule(
                    match_rules=(MatchRule(startswith="cli alias name wr "),)
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="aaa authentication login console"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(MatchRule(startswith="port-channel load-balance"),),
                ),
                IdempotentCommandsRule(
                    match_rules=(MatchRule(startswith="hostname "),)
                ),
                IdempotentCommandsRule(
                    match_rules=(MatchRule(startswith="ip tftp source-interface"),),
                ),
                IdempotentCommandsRule(
                    match_rules=(MatchRule(startswith="ip telnet source-interface"),),
                ),
                IdempotentCommandsRule(
                    match_rules=(MatchRule(startswith="ip tacacs source-interface"),),
                ),
                IdempotentCommandsRule(
                    match_rules=(MatchRule(startswith="logging source-interface"),),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="hardware access-list tcam region ifacl"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="hardware access-list tcam region vacl"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="hardware access-list tcam region qos"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="hardware access-list tcam region racl"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(
                            startswith="hardware access-list tcam region ipv6-racl"
                        ),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(
                            startswith="hardware access-list tcam region e-ipv6-racl"
                        ),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="hardware access-list tcam region l3qos"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="router ospf"),
                        MatchRule(startswith="vrf"),
                        MatchRule(startswith="maximum-paths"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="router ospf"),
                        MatchRule(startswith="vrf"),
                        MatchRule(startswith="log-adjacency-changes"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="router ospf"),
                        MatchRule(startswith="maximum-paths"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="router ospf"),
                        MatchRule(startswith="log-adjacency-changes"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="router bgp"),
                        MatchRule(startswith="vrf"),
                        MatchRule(startswith="address-family"),
                        MatchRule(startswith="maximum-paths"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="router bgp"),
                        MatchRule(startswith="address-family"),
                        MatchRule(startswith="maximum-paths"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="router bgp"),
                        MatchRule(startswith="template"),
                        MatchRule(startswith="address-family"),
                        MatchRule(startswith="send-community"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="interface"),
                        MatchRule(re_search="^hsrp \\d+"),
                        MatchRule(startswith="ip"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="interface"),
                        MatchRule(re_search="^hsrp \\d+"),
                        MatchRule(startswith="priority"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="interface"),
                        MatchRule(re_search="^hsrp \\d+"),
                        MatchRule(startswith="authentication md5 key-string"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="interface"),
                        MatchRule(startswith="ip address"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="interface"),
                        MatchRule(startswith="duplex"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="interface"),
                        MatchRule(startswith="speed"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="interface"),
                        MatchRule(startswith="switchport mode"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="interface"),
                        MatchRule(startswith="switchport access vlan"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="interface"),
                        MatchRule(startswith="switchport trunk native vlan"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="interface"),
                        MatchRule(startswith="switchport trunk allowed vlan"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="interface"),
                        MatchRule(startswith="udld port"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="interface"),
                        MatchRule(startswith="ip ospf cost"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="interface"),
                        MatchRule(startswith="ipv6 link-local"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="interface"),
                        MatchRule(startswith="ospfv3 cost"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="interface"),
                        MatchRule(startswith="mtu"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(equals="line console"),
                        MatchRule(startswith="exec-timeout"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="line vty"),
                        MatchRule(startswith="transport input"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="line vty"),
                        MatchRule(startswith="ipv6 access-class"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="line vty"),
                        MatchRule(startswith="access-class"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="router bgp"),
                        MatchRule(
                            startswith="bgp router-id",
                        ),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="router bgp"),
                        MatchRule(
                            re_search="neighbor \\S+ description",
                        ),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="router ospf"),
                        MatchRule(startswith="router-id"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="router ospf"),
                        MatchRule(startswith="log-adjacency-changes"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="ipv6 router ospf"),
                        MatchRule(startswith="router-id"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="ipv6 router ospf"),
                        MatchRule(startswith="log-adjacency-changes"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(MatchRule(startswith="mac address-table aging-time"),),
                ),
                IdempotentCommandsRule(
                    match_rules=(MatchRule(startswith="snmp-server location"),)
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="vpc domain"),
                        MatchRule(startswith="role priority"),
                    ),
                ),
                IdempotentCommandsRule(match_rules=(MatchRule(startswith="banner"),)),
                IdempotentCommandsRule(
                    match_rules=(MatchRule(startswith="username admin password 5"),),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(
                            equals="policy-map type control-plane copp-system-policy"
                        ),
                        MatchRule(startswith="class"),
                        MatchRule(startswith="police"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="router bgp"),
                        MatchRule(startswith="vrf"),
                        MatchRule(startswith="neighbor"),
                        MatchRule(startswith="address-family"),
                        MatchRule(startswith="soft-reconfiguration inbound"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="router bgp"),
                        MatchRule(startswith="vrf"),
                        MatchRule(startswith="neighbor"),
                        MatchRule(startswith="password"),
                    ),
                ),
            ],
            negation_default_when=[
                NegationDefaultWhenRule(
                    match_rules=(
                        MatchRule(startswith="interface"),
                        MatchRule(
                            startswith="ip ospf bfd",
                            re_search="standby \\d+ authentication md5 key-string",
                        ),
                    ),
                ),
                NegationDefaultWhenRule(
                    match_rules=(
                        MatchRule(startswith="router bgp"),
                        MatchRule(startswith="neighbor"),
                        MatchRule(startswith="address-family"),
                        MatchRule(equals="send-community"),
                    ),
                ),
                NegationDefaultWhenRule(
                    match_rules=(
                        MatchRule(startswith="interface"),
                        MatchRule(contains="ip ospf passive-interface"),
                    ),
                ),
                NegationDefaultWhenRule(
                    match_rules=(
                        MatchRule(startswith="interface"),
                        MatchRule(contains="ospfv3 passive-interface"),
                    ),
                ),
            ],
            negate_with=[
                NegationDefaultWithRule(
                    match_rules=(
                        MatchRule(startswith="router bgp"),
                        MatchRule(startswith="address-family"),
                        MatchRule(startswith="maximum-paths ibgp"),
                    ),
                    use="default maximum-paths ibgp",
                ),
                NegationDefaultWithRule(
                    match_rules=(
                        MatchRule(startswith="router bgp"),
                        MatchRule(startswith="vrf"),
                        MatchRule(startswith="address-family"),
                        MatchRule(startswith="maximum-paths ibgp"),
                    ),
                    use="default maximum-paths ibgp",
                ),
                NegationDefaultWithRule(
                    match_rules=(
                        MatchRule(equals="line vty"),
                        MatchRule(startswith="session-limit"),
                    ),
                    use="session-limit 32",
                ),
            ],
        )



from logging import getLogger

from hier_config.models import (
    IdempotentCommandsRule,
    MatchRule,
    NegationDefaultWithRule,
    OrderingRule,
    ParentAllowsDuplicateChildRule,
    PerLineSubRule,
    SectionalExitingRule,
)
from hier_config.platforms.driver_base import HConfigDriverBase, HConfigDriverRules
from hier_config.root import HConfig

logger = getLogger(__name__)


def _rm_ipv6_acl_sequence_numbers(config: HConfig) -> None:
    """If there are sequence numbers in the IPv6 ACL, remove them."""
    for acl in config.get_children(startswith="ipv6 access-list "):
        for entry in acl.children:
            if entry.text.startswith("sequence"):
                entry.text = " ".join(entry.text.split()[2:])


def _remove_ipv4_acl_remarks(config: HConfig) -> None:
    for acl in config.get_children(startswith="ip access-list "):
        for entry in tuple(acl.children):
            if entry.text.startswith("remark"):
                entry.delete()


def _add_acl_sequence_numbers(config: HConfig) -> None:
    """Add ACL sequence numbers."""
    ipv4_acl_sw = "ip access-list"
    acl_line_sw: tuple[str, ...] = ("permit", "deny")
    for child in config.children:
        if child.text.startswith(ipv4_acl_sw):
            sequence_number = 10
            for sub_child in child.children:
                if sub_child.text.startswith(acl_line_sw):
                    sub_child.text = f"{sequence_number} {sub_child.text}"
                    sequence_number += 10


class HConfigDriverCiscoIOS(HConfigDriverBase):
    """Driver for Cisco IOS and IOS-XE.

    Includes post-load callbacks that normalise IPv4/IPv6 ACL sequence numbers
    and remove IPv4 ACL remarks, ensuring stable diffs across device snapshots.
    Also handles BGP template peer-policy/peer-session sectional exiting and
    ``logging console`` negation-with replacement.  Platform enum:
    ``Platform.CISCO_IOS``.
    """

    @staticmethod
    def _instantiate_rules() -> HConfigDriverRules:
        return HConfigDriverRules(
            negate_with=[
                NegationDefaultWithRule(
                    match_rules=(MatchRule(startswith="logging console "),),
                    use="logging console debugging",
                ),
            ],
            sectional_exiting=[
                SectionalExitingRule(
                    match_rules=(
                        MatchRule(startswith="router bgp"),
                        MatchRule(startswith="template peer-policy"),
                    ),
                    exit_text="exit-peer-policy",
                ),
                SectionalExitingRule(
                    match_rules=(
                        MatchRule(startswith="router bgp"),
                        MatchRule(startswith="template peer-session"),
                    ),
                    exit_text="exit-peer-session",
                ),
                SectionalExitingRule(
                    match_rules=(
                        MatchRule(startswith="router bgp"),
                        MatchRule(startswith="address-family"),
                    ),
                    exit_text="exit-address-family",
                ),
            ],
            ordering=[
                OrderingRule(
                    match_rules=(
                        MatchRule(startswith="interface"),
                        MatchRule(startswith="switchport mode "),
                    ),
                    weight=-10,
                ),
                OrderingRule(
                    match_rules=(MatchRule(startswith="no vlan filter"),),
                    weight=200,
                ),
                OrderingRule(
                    match_rules=(
                        MatchRule(startswith="interface"),
                        MatchRule(startswith="no shutdown"),
                    ),
                    weight=200,
                ),
                OrderingRule(
                    match_rules=(
                        MatchRule(startswith="aaa group server tacacs+ "),
                        MatchRule(startswith="no server "),
                    ),
                    weight=10,
                ),
                OrderingRule(
                    match_rules=(MatchRule(startswith="no tacacs-server "),),
                    weight=10,
                ),
            ],
            parent_allows_duplicate_child=[
                ParentAllowsDuplicateChildRule(
                    match_rules=(
                        MatchRule(startswith="router"),
                        MatchRule(startswith="address-family"),
                    ),
                ),
            ],
            per_line_sub=[
                PerLineSubRule(search="^Building configuration.*", replace=""),
                PerLineSubRule(search="^Current configuration.*", replace=""),
                PerLineSubRule(search="^! Last configuration change.*", replace=""),
                PerLineSubRule(search="^! NVRAM config last updated.*", replace=""),
                PerLineSubRule(search="^ntp clock-period .*", replace=""),
                PerLineSubRule(search="^version.*", replace=""),
                PerLineSubRule(search="^ logging event link-status$", replace=""),
                PerLineSubRule(search="^ logging event subif-link-status$", replace=""),
                PerLineSubRule(search="^\\s*ipv6 unreachables disable$", replace=""),
                PerLineSubRule(search="^end$", replace=""),
                PerLineSubRule(search="^\\s*[#!].*", replace=""),
                PerLineSubRule(search="^ no ip address", replace=""),
                PerLineSubRule(search="^ exit-peer-policy", replace=""),
                PerLineSubRule(search="^ exit-peer-session", replace=""),
                PerLineSubRule(search="^ exit-address-family", replace=""),
                PerLineSubRule(
                    search="^crypto key generate rsa general-keys.*$", replace=""
                ),
            ],
            idempotent_commands=[
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="vlan"),
                        MatchRule(startswith="name"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="interface "),
                        MatchRule(startswith="description "),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="interface "),
                        MatchRule(startswith="ip address "),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="interface "),
                        MatchRule(startswith="switchport mode "),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="interface "),
                        MatchRule(startswith="authentication host-mode "),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="interface "),
                        MatchRule(
                            startswith="authentication event server dead action authorize vlan ",
                        ),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="errdisable recovery interval "),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(MatchRule(re_search=r"^(no )?logging console.*"),),
                ),
            ],
            post_load_callbacks=[
                _rm_ipv6_acl_sequence_numbers,
                _remove_ipv4_acl_remarks,
                _add_acl_sequence_numbers,
            ],
        )



from collections.abc import Iterable

from hier_config.child import HConfigChild
from hier_config.models import (
    MatchRule,
    ParentAllowsDuplicateChildRule,
    PerLineSubRule,
    SectionalExitingRule,
)
from hier_config.platforms.driver_base import HConfigDriverBase, HConfigDriverRules


class HConfigDriverFortinetFortiOS(HConfigDriverBase):
    """Driver for Fortinet FortiOS."""

    @staticmethod
    def _instantiate_rules() -> HConfigDriverRules:
        return HConfigDriverRules(
            sectional_exiting=[
                SectionalExitingRule(
                    match_rules=(MatchRule(startswith="config "),), exit_text="end"
                ),
                SectionalExitingRule(
                    match_rules=(
                        MatchRule(startswith="config "),
                        MatchRule(startswith="edit "),
                    ),
                    exit_text="next",
                ),
            ],
            parent_allows_duplicate_child=[
                ParentAllowsDuplicateChildRule(
                    match_rules=(MatchRule(startswith="config"),)
                ),
            ],
            per_line_sub=[
                PerLineSubRule(search="^end$", replace=" end"),
                PerLineSubRule(search="^next$", replace="  next"),
            ],
        )

    def swap_negation(self, child: HConfigChild) -> HConfigChild:
        """Swap negation of a `self.text`."""
        if child.text.startswith(self.negation_prefix):
            child.text = f"{self.declaration_prefix}{child.text_without_negation}"
        elif child.text.startswith(self.declaration_prefix):
            child.text = f"{self.negation_prefix}{child.text.removeprefix(self.declaration_prefix).split()[0]}"

        return child

    def idempotent_for(
        self, config: HConfigChild, other_children: Iterable[HConfigChild]
    ) -> HConfigChild | None:
        """Override idempotent_for to only consider a config idempotent
        if the same command exists in the other set.
        """
        for other_child in other_children:
            if (
                config.text.startswith(self.declaration_prefix)
                and other_child.text.startswith(self.declaration_prefix)
                and config.text.split()[1] == other_child.text.split()[1]
            ):
                return other_child
        return super().idempotent_for(config, other_children)

    @property
    def negation_prefix(self) -> str:
        return "unset "

    @property
    def declaration_prefix(self) -> str:
        return "set "



from __future__ import annotations

from abc import ABC, abstractmethod
from itertools import chain
from logging import getLogger
from typing import TYPE_CHECKING, TypeVar

from .children import HConfigChildren
from .exceptions import DuplicateChildError

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator

    from .child import HConfigChild
    from .models import MatchRule, SetLikeOfStr
    from .platforms.driver_base import HConfigDriverBase
    from .root import HConfig

    _HConfigRootOrChildT = TypeVar("_HConfigRootOrChildT", bound=HConfig | HConfigChild)

logger = getLogger(__name__)


class HConfigBase(ABC):  # noqa: PLR0904
    """Abstract base class for the hierarchical configuration tree.

    Both `HConfig` (the root) and `HConfigChild` (individual nodes) inherit from
    this class.  It provides the shared tree-manipulation API: adding, searching,
    and diffing children, as well as the `_future` / `_config_to_get_to` algorithms
    that power `WorkflowRemediation`.
    """

    __slots__ = ("children",)

    def __init__(self) -> None:
        self.children = HConfigChildren()

    def __len__(self) -> int:
        return len(tuple(self.all_children()))

    def __bool__(self) -> bool:
        return True

    @abstractmethod
    def __hash__(self) -> int:
        pass

    def __iter__(self) -> Iterator[HConfigChild]:
        return iter(self.children)

    @property
    @abstractmethod
    def root(self) -> HConfig:
        pass

    @property
    @abstractmethod
    def driver(self) -> HConfigDriverBase:
        pass

    @abstractmethod
    def lineage(self) -> Iterator[HConfigChild]:
        pass

    @abstractmethod
    def depth(self) -> int:
        pass

    def add_children(self, lines: Iterable[str]) -> None:
        """Add child instances of HConfigChild."""
        for line in lines:
            self.add_child(line)

    def add_child(
        self,
        text: str,
        *,
        return_if_present: bool = False,
        check_if_present: bool = True,
    ) -> HConfigChild:
        """Add a child instance of HConfigChild."""
        if not text:
            message = "text was empty"
            raise ValueError(message)

        if check_if_present and (child := self.children.get(text)):
            if self._is_duplicate_child_allowed():
                new_child = self.instantiate_child(text)
                self.children.append(new_child, update_mapping=False)
                return new_child
            if return_if_present:
                return child
            message = f"Found a duplicate section: {(*self.path(), text)}"
            raise DuplicateChildError(message)

        new_child = self.instantiate_child(text)
        self.children.append(new_child)
        return new_child

    def path(self) -> Iterator[str]:  # noqa: PLR6301
        yield from ()

    def add_deep_copy_of(
        self,
        child_to_add: HConfigChild,
        *,
        merged: bool = False,
    ) -> HConfigChild:
        """Add a nested copy of a child to self."""
        new_child = self.add_shallow_copy_of(child_to_add, merged=merged)
        for child in child_to_add.children:
            new_child.add_deep_copy_of(child, merged=merged)

        return new_child

    def all_children_sorted(self) -> Iterator[HConfigChild]:
        """Recursively find and yield all children sorted at each hierarchy."""
        for child in sorted(self.children):
            yield child
            yield from child.all_children_sorted()

    def all_children(self) -> Iterator[HConfigChild]:
        """Recursively find and yield all children at each hierarchy."""
        for child in self.children:
            yield child
            yield from child.all_children()

    def get_child_deep(self, match_rules: tuple[MatchRule, ...]) -> HConfigChild | None:
        """Find the first child recursively given a tuple of MatchRules."""
        return next(self.get_children_deep(match_rules), None)

    def get_children_deep(
        self,
        match_rules: tuple[MatchRule, ...],
    ) -> Iterator[HConfigChild]:
        """Find children recursively given a tuple of MatchRules."""
        rule = match_rules[0]
        remaining_rules = match_rules[1:]
        for child in self.get_children(
            equals=rule.equals,
            startswith=rule.startswith,
            endswith=rule.endswith,
            contains=rule.contains,
            re_search=rule.re_search,
        ):
            if remaining_rules:
                yield from child.get_children_deep(remaining_rules)
            else:
                yield child

    def get_child(
        self,
        *,
        equals: str | SetLikeOfStr | None = None,
        startswith: str | tuple[str, ...] | None = None,
        endswith: str | tuple[str, ...] | None = None,
        contains: str | tuple[str, ...] | None = None,
        re_search: str | None = None,
    ) -> HConfigChild | None:
        """Find a child by text_match rule. If it is not found, return None."""
        return next(
            self.get_children(
                equals=equals,
                startswith=startswith,
                endswith=endswith,
                contains=contains,
                re_search=re_search,
            ),
            None,
        )

    def get_children(
        self,
        *,
        equals: str | SetLikeOfStr | None = None,
        startswith: str | tuple[str, ...] | None = None,
        endswith: str | tuple[str, ...] | None = None,
        contains: str | tuple[str, ...] | None = None,
        re_search: str | None = None,
    ) -> Iterator[HConfigChild]:
        """Find all children matching a text_match rule and return them."""
        # For isinstance(equals, str) only matches, find the first child using children_dict
        children_slice = slice(None, None)
        if (
            isinstance(equals, str)
            and startswith is endswith is contains is re_search is None
        ):
            if child := self.children.get(equals):
                yield child
                children_slice = slice(self.children.index(child) + 1, None)
            else:
                return

        elif (
            isinstance(startswith, (str, tuple))
            and equals is endswith is contains is re_search is None
        ):
            duplicates_allowed = None
            for index, child in enumerate(self.children):
                if child.text.startswith(startswith):
                    yield child
                    if duplicates_allowed is None:
                        duplicates_allowed = self._is_duplicate_child_allowed()
                    if duplicates_allowed:
                        children_slice = slice(index + 1, None)
                        break
            else:
                return

        for child in self.children[children_slice]:
            if child.is_match(
                equals=equals,
                startswith=startswith,
                endswith=endswith,
                contains=contains,
                re_search=re_search,
            ):
                yield child

    def add_shallow_copy_of(
        self,
        child_to_add: HConfigChild,
        *,
        merged: bool = False,
    ) -> HConfigChild:
        """Add a nested copy of a child_to_add to self.children."""
        new_child = self.add_child(child_to_add.text, return_if_present=merged)

        if merged:
            new_child.instances.append(child_to_add.instance)
        new_child.comments.update(child_to_add.comments)
        new_child.order_weight = child_to_add.order_weight
        if child_to_add.is_leaf:
            new_child.tags_add(child_to_add.tags)

        return new_child

    def unified_diff(self, target: HConfig | HConfigChild) -> Iterator[str]:
        """Yield unified-diff lines comparing self to target.

        Each yielded string is prefixed with ``-`` (present in self but not
        target) or ``+`` (present in target but not self), followed by the
        appropriate indentation and the command text.

        .. note::
            This algorithm does not account for duplicate child differences
            (e.g. two ``endif`` tokens in an IOS-XR route-policy) and does
            not preserve command order where it matters (e.g. ACLs without
            sequence numbers).  Use sequence numbers in ACL entries when
            order is significant.

        Produces output similar to :func:`difflib.unified_diff`.
        """
        # if a self child is missing from the target "- self_child.text"
        for self_child in self.children:
            self_iter = iter((f"{self_child.indentation}{self_child.text}",))
            if target_child := target.children.get(self_child.text, None):
                found = self_child.unified_diff(target_child)
                if peek := next(found, None):
                    yield from chain(self_iter, (peek,), found)
            else:
                yield f"{self_child.indentation}- {self_child.text}"
                yield from (
                    f"{c.indentation}- {c.text}"
                    for c in self_child.all_children_sorted()
                )
        # if a target child is missing from self "+ target_child.text"
        for target_child in target.children:
            if target_child.text not in self.children:
                yield f"{target_child.indentation}+ {target_child.text}"
                yield from (
                    f"{c.indentation}+ {c.text}"
                    for c in target_child.all_children_sorted()
                )

    def _future_pre(self, config: HConfig | HConfigChild) -> tuple[set[str], set[str]]:
        negated_or_recursed: set[str] = set()
        config_children_ignore: set[str] = set()
        for self_child in self.children:
            # Is the command effectively negating a command in self.children?
            if (negation_text := self.root.driver.negate_with(self_child)) and (
                config_child := config.get_child(equals=negation_text)
            ):
                negated_or_recursed.add(self_child.text)
                config_children_ignore.add(config_child.text)
        return negated_or_recursed, config_children_ignore

    def _future(  # noqa: C901
        self,
        config: HConfig | HConfigChild,
        future_config: HConfig | HConfigChild,
    ) -> None:
        """Recursively compute the future configuration subtree.

        Called by :meth:`HConfig.future` to walk the config tree and merge
        ``config`` on top of ``self``, applying driver-specific rules for
        sectional overwrite, idempotency, and negation.  The result is written
        into ``future_config``.

        Known gaps (not yet accounted for):

        - Negating a numbered ACL when removing a single entry
        - Idempotent command avoid list
        - And likely other edge cases
        """
        negated_or_recursed, config_children_ignore = self._future_pre(config)

        for config_child in config.children:
            if config_child.text in config_children_ignore:
                continue
            # sectional_overwrite
            # sectional_overwrite_no_negate
            if (
                config_child.use_sectional_overwrite()
                or config_child.use_sectional_overwrite_without_negation()
            ):
                future_config.add_deep_copy_of(config_child)
            # Idempotent commands
            elif self_child := self.root.driver.idempotent_for(
                config_child,
                self.children,
            ):
                future_config.add_deep_copy_of(config_child)
                negated_or_recursed.add(self_child.text)
            # config_child is already in self
            elif self_child := self.get_child(equals=config_child.text):
                future_child = future_config.add_shallow_copy_of(self_child)
                self_child._future(config_child, future_child)  # noqa: SLF001
                negated_or_recursed.add(config_child.text)
            # config_child is being negated
            elif config_child.text.startswith(self.driver.negation_prefix):
                unnegated_command = config_child.text_without_negation
                if self.get_child(equals=unnegated_command):
                    negated_or_recursed.add(unnegated_command)
                # Account for "no ..." commands in the running config
                else:
                    future_config.add_shallow_copy_of(config_child)
            # The negated form of config_child is in self.children
            elif self_child := self.get_child(
                equals=f"{self.driver.negation_prefix}{config_child.text}",
            ):
                negated_or_recursed.add(self_child.text)
            # config_child is not in self and doesn't match a special case
            else:
                future_config.add_deep_copy_of(config_child)

        for self_child in self.children:
            # self_child matched an above special case and should be ignored
            if self_child.text in negated_or_recursed:
                continue
            # self_child was not modified above and should be present in the future config
            future_config.add_deep_copy_of(self_child)

    @abstractmethod
    def instantiate_child(self, text: str) -> HConfigChild:
        pass

    @abstractmethod
    def _is_duplicate_child_allowed(self) -> bool:
        pass

    def _with_tags(
        self,
        tags: frozenset[str],
        new_instance: _HConfigRootOrChildT,
    ) -> _HConfigRootOrChildT:
        """Adds children recursively that have a subset of tags."""
        for child in self.children:
            if tags.issubset(child.tags):
                new_child = new_instance.add_shallow_copy_of(child)
                child._with_tags(tags, new_instance=new_child)  # noqa: SLF001

        return new_instance

    def _config_to_get_to(
        self,
        target: _HConfigRootOrChildT,
        delta: _HConfigRootOrChildT,
    ) -> _HConfigRootOrChildT:
        """Figures out what commands need to be executed to transition from self to target.
        self is the source data structure(i.e. the running_config),
        target is the destination(i.e. generated_config).

        """
        self._config_to_get_to_left(target, delta)
        self._config_to_get_to_right(target, delta)

        return delta

    @staticmethod
    def _strip_acl_sequence_number(hier_child: HConfigChild) -> str:
        words = hier_child.text.split()
        if words[0].isdecimal():
            words.pop(0)
        return " ".join(words)

    def _difference(
        self,
        target: _HConfigRootOrChildT,
        delta: _HConfigRootOrChildT,
        target_acl_children: dict[str, HConfigChild] | None = None,
        *,
        in_acl: bool = False,
    ) -> _HConfigRootOrChildT:
        acl_sw_matches = tuple(f"ip{x} access-list " for x in ("", "v4", "v6"))

        for self_child in self.children:
            # Not dealing with negations and defaults for now
            if self_child.text.startswith((self.driver.negation_prefix, "default ")):
                continue

            if in_acl:
                # Ignore ACL sequence numbers
                if target_acl_children is None:
                    message = "target_acl_children cannot be None"
                    raise TypeError(message)
                target_child = target_acl_children.get(
                    self._strip_acl_sequence_number(self_child),
                )
            else:
                target_child = target.get_child(equals=self_child.text)

            if target_child is None:
                delta.add_deep_copy_of(self_child)
            else:
                delta_child = delta.add_child(self_child.text)
                if self_child.text.startswith(acl_sw_matches):
                    self_child._difference(  # noqa: SLF001
                        target_child,
                        delta_child,
                        target_acl_children={
                            self._strip_acl_sequence_number(c): c
                            for c in target_child.children
                        },
                        in_acl=True,
                    )
                else:
                    self_child._difference(target_child, delta_child)  # noqa: SLF001
                if not delta_child.children:
                    delta_child.delete()

        return delta

    def _config_to_get_to_left(
        self,
        target: HConfig | HConfigChild,
        delta: HConfig | HConfigChild,
    ) -> None:
        # find self.children that are not in target.children
        # i.e. what needs to be negated or defaulted
        # Also, find out if another command in self.children will overwrite
        # i.e. be idempotent
        for self_child in self.children:
            if self_child.text in target.children:
                continue
            if self_child.is_idempotent_command(target.children):
                continue

            # in other but not self
            # add this node but not any children
            negated = delta.add_child(self_child.text).negate()
            if self_child.children:
                negated.comments.add(f"removes {len(self_child.children) + 1} lines")

    def _config_to_get_to_right(
        self,
        target: HConfig | HConfigChild,
        delta: HConfig | HConfigChild,
    ) -> None:
        # Find what would need to be added to source_config to get to self
        for target_child in target.children:
            # If the child exist, recurse into its children
            if self_child := self.children.get(target_child.text):
                # Do we need to rewrite the child and its children as well?
                if self_child.use_sectional_overwrite():
                    self_child.overwrite_with(target_child, delta)
                    continue
                if self_child.use_sectional_overwrite_without_negation():
                    self_child.overwrite_with(target_child, delta, negate=False)
                    continue
                # This creates a new HConfigChild object just in case there are some delta children.
                # This is not very efficient, think of a way to not do this.
                subtree = delta.instantiate_child(target_child.text)
                self_child._config_to_get_to(target_child, subtree)  # noqa: SLF001
                if subtree.children:
                    delta.children.append(subtree)
            # The child is absent, add it.
            else:
                # If the target_child is already in the delta, that means it was negated in the target config
                if target_child.text in delta.children:
                    continue
                new_item = delta.add_deep_copy_of(target_child)
                # Mark the new item and all of its children as new_in_config.
                new_item.new_in_config = True
                for child in new_item.all_children():
                    child.new_in_config = True
                if new_item.children:
                    new_item.comments.add("new section")



from __future__ import annotations

from typing import TYPE_CHECKING

from hier_config.child import HConfigChild
from hier_config.models import (
    IdempotentCommandsRule,
    IndentAdjustRule,
    MatchRule,
    OrderingRule,
    ParentAllowsDuplicateChildRule,
    PerLineSubRule,
    SectionalExitingRule,
    SectionalOverwriteNoNegateRule,
    SectionalOverwriteRule,
)
from hier_config.platforms.driver_base import HConfigDriverBase, HConfigDriverRules

if TYPE_CHECKING:
    from collections.abc import Iterable

    from hier_config.root import HConfig


def _fixup_xr_comments(config: HConfig) -> None:
    """Move ``!`` comment lines into the next sibling's comments set."""
    for parent in (config, *config.all_children()):
        siblings = list(parent.children)
        comment_buffer: list[str] = []
        for sibling in siblings:
            if sibling.text.startswith("!"):
                comment_text = sibling.text.removeprefix("!").lstrip(" ")
                if comment_text:
                    comment_buffer.append(comment_text)
                sibling.delete()
            elif comment_buffer:
                for comment in comment_buffer:
                    sibling.comments.add(comment)
                comment_buffer.clear()


class HConfigDriverCiscoIOSXR(HConfigDriverBase):  # pylint: disable=too-many-instance-attributes
    """Driver for Cisco IOS XR.

    Extends the base idempotency logic to handle XR-style ACL sequence numbers
    (``ipv4 access-list`` / ``ipv6 access-list``).  Also configures XR-specific
    sectional exiting (``end-policy``, ``end-set``, ``end-template``, ``root``),
    sectional overwrite for templates, indent-adjust for inline templates, and
    duplicate-child allowances inside ``route-policy`` blocks.
    Platform enum: ``Platform.CISCO_XR``.
    """

    def idempotent_for(
        self,
        config: HConfigChild,
        other_children: Iterable[HConfigChild],
    ) -> HConfigChild | None:
        if isinstance(config.parent, HConfigChild):
            acl = ("ipv4 access-list ", "ipv6 access-list ")
            if config.parent.text.startswith(acl):
                self_sn = config.text.split(" ", 1)[0]
                for other_child in other_children:
                    other_sn = other_child.text.split(" ", 1)[0]
                    if self_sn == other_sn:
                        return other_child

        return super().idempotent_for(config, other_children)

    @staticmethod
    def _instantiate_rules() -> HConfigDriverRules:
        return HConfigDriverRules(
            sectional_exiting=[
                SectionalExitingRule(
                    match_rules=(MatchRule(startswith="route-policy"),),
                    exit_text="end-policy",
                    exit_text_parent_level=True,
                ),
                SectionalExitingRule(
                    match_rules=(MatchRule(startswith="prefix-set"),),
                    exit_text="end-set",
                    exit_text_parent_level=True,
                ),
                SectionalExitingRule(
                    match_rules=(MatchRule(startswith="policy-map"),),
                    exit_text="end-policy-map",
                    exit_text_parent_level=True,
                ),
                SectionalExitingRule(
                    match_rules=(MatchRule(startswith="class-map"),),
                    exit_text="end-class-map",
                    exit_text_parent_level=True,
                ),
                SectionalExitingRule(
                    match_rules=(MatchRule(startswith="community-set"),),
                    exit_text="end-set",
                    exit_text_parent_level=True,
                ),
                SectionalExitingRule(
                    match_rules=(MatchRule(startswith="extcommunity-set"),),
                    exit_text="end-set",
                    exit_text_parent_level=True,
                ),
                SectionalExitingRule(
                    match_rules=(MatchRule(startswith="template"),),
                    exit_text="end-template",
                    exit_text_parent_level=True,
                ),
                SectionalExitingRule(
                    match_rules=(MatchRule(startswith="group"),),
                    exit_text="end-group",
                    exit_text_parent_level=True,
                ),
                SectionalExitingRule(
                    match_rules=(MatchRule(startswith="interface"),),
                    exit_text="root",
                ),
                SectionalExitingRule(
                    match_rules=(MatchRule(startswith="router bgp"),),
                    exit_text="root",
                ),
            ],
            sectional_overwrite=[
                SectionalOverwriteRule(match_rules=(MatchRule(startswith="template"),)),
            ],
            sectional_overwrite_no_negate=[
                SectionalOverwriteNoNegateRule(
                    match_rules=(MatchRule(startswith="as-path-set"),)
                ),
                SectionalOverwriteNoNegateRule(
                    match_rules=(MatchRule(startswith="prefix-set"),)
                ),
                SectionalOverwriteNoNegateRule(
                    match_rules=(MatchRule(startswith="route-policy"),)
                ),
                SectionalOverwriteNoNegateRule(
                    match_rules=(MatchRule(startswith="extcommunity-set"),),
                ),
                SectionalOverwriteNoNegateRule(
                    match_rules=(MatchRule(startswith="community-set"),),
                ),
            ],
            ordering=[
                OrderingRule(
                    match_rules=(MatchRule(startswith="vrf "),),
                    weight=-200,
                ),
                OrderingRule(
                    match_rules=(MatchRule(startswith="no vrf "),),
                    weight=200,
                ),
            ],
            indent_adjust=[
                IndentAdjustRule(
                    start_expression="^\\s*template(?!\\s+timeout)",
                    end_expression="^\\s*end-template",
                ),
            ],
            parent_allows_duplicate_child=[
                ParentAllowsDuplicateChildRule(
                    match_rules=(MatchRule(startswith="route-policy"),)
                ),
                ParentAllowsDuplicateChildRule(
                    match_rules=(
                        MatchRule(startswith="route-policy"),
                        MatchRule(startswith="if "),
                    )
                ),
            ],
            per_line_sub=[
                PerLineSubRule(search="^Building configuration.*", replace=""),
                PerLineSubRule(search="^Current configuration.*", replace=""),
                PerLineSubRule(search="^ntp clock-period .*", replace=""),
                PerLineSubRule(search=".*speed.*", replace=""),
                PerLineSubRule(search=".*duplex.*", replace=""),
                PerLineSubRule(search=".*negotiation auto.*", replace=""),
                PerLineSubRule(search=".*parity none.*", replace=""),
                PerLineSubRule(search="^end-policy$", replace=" end-policy"),
                PerLineSubRule(search="^end-set$", replace=" end-set"),
                PerLineSubRule(search="^end-group$", replace=" end-group"),
                PerLineSubRule(search="^end$", replace=""),
                PerLineSubRule(search="^\\s*#.*", replace=""),
                PerLineSubRule(search="^\\s*!\\s*$", replace=""),
            ],
            post_load_callbacks=[_fixup_xr_comments],
            idempotent_commands=[
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="router bgp"),
                        MatchRule(startswith="vrf"),
                        MatchRule(startswith="address-family"),
                        MatchRule(startswith="additional-paths selection route-policy"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="router bgp"),
                        MatchRule(startswith="bgp router-id"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="router bgp"),
                        MatchRule(startswith="neighbor-group"),
                        MatchRule(startswith="address-family"),
                        MatchRule(startswith="soft-reconfiguration inbound"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="router bgp"),
                        MatchRule(startswith="vrf"),
                        MatchRule(startswith="neighbor"),
                        MatchRule(startswith="address-family"),
                        MatchRule(startswith="soft-reconfiguration inbound"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="router bgp"),
                        MatchRule(startswith="vrf"),
                        MatchRule(startswith="neighbor"),
                        MatchRule(startswith="address-family"),
                        MatchRule(startswith="maximum-prefix"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="router bgp"),
                        MatchRule(startswith="vrf"),
                        MatchRule(startswith="neighbor"),
                        MatchRule(startswith="password"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="router bgp"),
                        MatchRule(startswith="vrf"),
                        MatchRule(startswith="neighbor"),
                        MatchRule(startswith="description"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="router bgp"),
                        MatchRule(startswith="neighbor"),
                        MatchRule(startswith="description"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="router bgp"),
                        MatchRule(startswith="neighbor"),
                        MatchRule(startswith="password"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="router ospf"),
                        MatchRule(startswith="area"),
                        MatchRule(startswith="interface"),
                        MatchRule(startswith="cost"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="router ospf"),
                        MatchRule(startswith="router-id"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="router ospf"),
                        MatchRule(startswith="area"),
                        MatchRule(startswith="message-digest-key"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="router ospf"),
                        MatchRule(startswith="max-metric router-lsa"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(equals="l2vpn"),
                        MatchRule(startswith="router-id"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(re_search="logging \\d+.\\d+.\\d+.\\d+ vrf MGMT"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(equals="line default"),
                        MatchRule(startswith="access-class ingress"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(equals="line default"),
                        MatchRule(startswith="transport input"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(MatchRule(startswith="hostname"),),
                ),
                IdempotentCommandsRule(
                    match_rules=(MatchRule(startswith="logging source-interface"),),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="interface"),
                        MatchRule(startswith="ipv4 address"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(MatchRule(startswith="snmp-server location"),),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(equals="line console"),
                        MatchRule(startswith="exec-timeout"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(equals="mpls ldp"),
                        MatchRule(startswith="session protection duration"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(equals="mpls ldp"),
                        MatchRule(startswith="igp sync delay"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="interface"),
                        MatchRule(startswith="mtu"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(MatchRule(startswith="banner"),),
                ),
            ],
        )



from hier_config.platforms.driver_base import HConfigDriverBase, HConfigDriverRules


class HConfigDriverHPComware5(HConfigDriverBase):
    """Driver for HP/H3C Comware 5 operating system.

    Overrides the negation prefix to ``"undo "`` to match Comware / H3C CLI
    conventions (e.g. ``undo ip address``).  No additional platform-specific
    rules are configured by default.  Platform enum: ``Platform.HP_COMWARE5``.
    """

    @property
    def negation_prefix(self) -> str:
        return "undo "

    @staticmethod
    def _instantiate_rules() -> HConfigDriverRules:
        return HConfigDriverRules()



from hier_config.models import (
    IdempotentCommandsRule,
    MatchRule,
    NegationDefaultWhenRule,
    PerLineSubRule,
    SectionalExitingRule,
)
from hier_config.platforms.driver_base import HConfigDriverBase, HConfigDriverRules


class HConfigDriverAristaEOS(HConfigDriverBase):
    """Driver for Arista EOS (Extensible Operating System).

    Handles EOS-specific conventions including BGP address-family sectional
    exiting, per-line substitutions that strip timestamps and banner markers,
    and a comprehensive set of idempotency rules for interface, BGP, and OSPF
    commands.  Platform enum: ``Platform.ARISTA_EOS``.
    """

    @staticmethod
    def _instantiate_rules() -> HConfigDriverRules:
        return HConfigDriverRules(
            sectional_exiting=[
                SectionalExitingRule(
                    match_rules=(
                        MatchRule(startswith="router bgp"),
                        MatchRule(startswith="template peer-policy"),
                    ),
                    exit_text="exit-peer-policy",
                ),
                SectionalExitingRule(
                    match_rules=(
                        MatchRule(startswith="router bgp"),
                        MatchRule(startswith="template peer-session"),
                    ),
                    exit_text="exit-peer-session",
                ),
                SectionalExitingRule(
                    match_rules=(
                        MatchRule(startswith="router bgp"),
                        MatchRule(startswith="address-family"),
                    ),
                    exit_text="exit-address-family",
                ),
            ],
            per_line_sub=[
                PerLineSubRule(search="^Building configuration.*", replace=""),
                PerLineSubRule(search="^Current configuration.*", replace=""),
                PerLineSubRule(search="^! Last configuration change.*", replace=""),
                PerLineSubRule(search="^! NVRAM config last updated.*", replace=""),
                PerLineSubRule(search="^ntp clock-period .*", replace=""),
                PerLineSubRule(search="^version.*", replace=""),
                PerLineSubRule(search="^ logging event link-status$", replace=""),
                PerLineSubRule(search="^ logging event subif-link-status$", replace=""),
                PerLineSubRule(search="^\\s*ipv6 unreachables disable$", replace=""),
                PerLineSubRule(search="^end$", replace=""),
                PerLineSubRule(search="^\\s*[#!].*", replace=""),
                PerLineSubRule(search="^ no ip address", replace=""),
                PerLineSubRule(search="^ exit-peer-policy", replace=""),
                PerLineSubRule(search="^ exit-peer-session", replace=""),
                PerLineSubRule(search="^ exit-address-family", replace=""),
            ],
            idempotent_commands=[
                IdempotentCommandsRule(
                    match_rules=(MatchRule(startswith="hostname"),),
                ),
                IdempotentCommandsRule(
                    match_rules=(MatchRule(startswith="logging source-interface"),),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="interface"),
                        MatchRule(startswith="ip address"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="line vty"),
                        MatchRule(
                            startswith="transport input",
                        ),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="line vty"),
                        MatchRule(
                            startswith="access-class",
                        ),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="line vty"),
                        MatchRule(
                            startswith="ipv6 access-class",
                        ),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="interface"),
                        MatchRule(
                            re_search="standby \\d+ (priority|authentication md5)",
                        ),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="router bgp"),
                        MatchRule(startswith="bgp router-id"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="router ospf"),
                        MatchRule(startswith="router-id"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="router ospf"),
                        MatchRule(startswith="max-lsa"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="router ospf"),
                        MatchRule(startswith="maximum-paths"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="ipv6 router ospf"),
                        MatchRule(startswith="router-id"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="router ospf"),
                        MatchRule(startswith="log-adjacency-changes"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="ipv6 router ospf"),
                        MatchRule(startswith="log-adjacency-changes"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="router bgp"),
                        MatchRule(re_search="neighbor \\S+ description"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(MatchRule(startswith="snmp-server location"),),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(equals="line con 0"),
                        MatchRule(startswith="exec-timeout"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="interface"),
                        MatchRule(startswith="ip ospf message-digest-key"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(MatchRule(startswith="logging buffered"),),
                ),
                IdempotentCommandsRule(
                    match_rules=(MatchRule(startswith="tacacs-server key"),),
                ),
                IdempotentCommandsRule(
                    match_rules=(MatchRule(startswith="logging facility"),),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="vlan internal allocation policy"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(MatchRule(startswith="username admin"),),
                ),
                IdempotentCommandsRule(
                    match_rules=(MatchRule(startswith="snmp-server user"),),
                ),
                IdempotentCommandsRule(
                    match_rules=(MatchRule(startswith="banner"),),
                ),
                IdempotentCommandsRule(
                    match_rules=(MatchRule(startswith="ntp source"),),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="management"),
                        MatchRule(startswith="idle-timeout"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(
                            startswith="aaa authentication enable default group tacacs+"
                        ),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(equals="control-plane"),
                        MatchRule(equals="ip access-group CPP in"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="interface"),
                        MatchRule(startswith="mtu"),
                    ),
                ),
                IdempotentCommandsRule(
                    match_rules=(MatchRule(startswith="snmp-server source-interface"),),
                ),
                IdempotentCommandsRule(
                    match_rules=(
                        MatchRule(startswith="ip tftp client source-interface"),
                    ),
                ),
            ],
            negation_default_when=[
                NegationDefaultWhenRule(
                    match_rules=(
                        MatchRule(startswith="interface"),
                        MatchRule(equals="logging event link-status"),
                    ),
                ),
            ],
        )



from .child import HConfigChild
from .constructors import (
    get_hconfig,
    get_hconfig_driver,
    get_hconfig_fast_load,
    get_hconfig_from_dump,
    get_hconfig_view,
)
from .models import ChangeDetail, MatchRule, Platform, ReportSummary, TagRule
from .reporting import RemediationReporter
from .root import HConfig
from .workflows import WorkflowRemediation

__all__ = (
    "ChangeDetail",
    "HConfig",
    "HConfigChild",
    "MatchRule",
    "Platform",
    "RemediationReporter",
    "ReportSummary",
    "TagRule",
    "WorkflowRemediation",
    "get_hconfig",
    "get_hconfig_driver",
    "get_hconfig_fast_load",
    "get_hconfig_from_dump",
    "get_hconfig_view",
)



from __future__ import annotations

from itertools import chain
from logging import getLogger
from re import search, sub
from typing import TYPE_CHECKING, Any

from .base import HConfigBase
from .models import Instance, MatchRule, SetLikeOfStr

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator

    from hier_config.platforms.driver_base import HConfigDriverBase

    from .root import HConfig


logger = getLogger(__name__)


class HConfigChild(  # noqa: PLR0904  pylint: disable=too-many-instance-attributes
    HConfigBase,
):
    """A single node in the hierarchical configuration tree.

    Each `HConfigChild` holds one configuration line (`text`), an ordered
    collection of its own children, optional tags/comments, and a reference
    back to its parent.  The tree is rooted at an `HConfig` instance; every
    other node is an `HConfigChild`.
    """

    __slots__ = (
        "_tags",
        "_text",
        "comments",
        "facts",
        "instances",
        "new_in_config",
        "order_weight",
        "parent",
        "real_indent_level",
    )

    def __init__(self, parent: HConfig | HConfigChild, text: str) -> None:
        super().__init__()
        self.parent = parent
        self._text: str = text.strip()
        self.real_indent_level: int
        # 0 is the default. Positive weights sink while negative weights rise.
        self.order_weight: int = 0
        self._tags: set[str] = set()
        self.comments: set[str] = set()
        self.new_in_config: bool = False
        self.instances: list[Instance] = []
        # To store externally inserted facts
        self.facts: dict[Any, Any] = {}

    def __str__(self) -> str:
        return "\n".join(self.lines(sectional_exiting=True))

    def __repr__(self) -> str:
        return f"HConfigChild(HConfig{'' if self.parent is self.root else 'Child'}, {self.text})"

    def __lt__(self, other: HConfigChild) -> bool:
        return self.order_weight < other.order_weight

    def __hash__(self) -> int:
        return hash(
            (
                self.text,
                # self.tags,
                # self.comments,
                self.new_in_config,
                self.order_weight,
                *self.children,
            ),
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, HConfigChild):
            return NotImplemented

        # We are intentionally not including the
        # comments, facts, instances, new_in_config, order_weight attributes.
        if self.text != other.text or self.tags != other.tags:
            return False

        return self.children == other.children

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    @property
    def driver(self) -> HConfigDriverBase:
        return self.root.driver

    @property
    def text(self) -> str:
        return self._text

    @text.setter
    def text(self, value: str) -> None:
        """Used for when self.text is changed after the object
        is instantiated to rebuild the children dictionary.
        """
        self._text = value.strip()
        self.parent.children.rebuild_mapping()

    @property
    def text_without_negation(self) -> str:
        return self.text.removeprefix(self.driver.negation_prefix)

    @property
    def root(self) -> HConfig:
        """Returns the HConfig object at the base of the tree."""
        return self.parent.root

    def lines(self, *, sectional_exiting: bool = False) -> Iterable[str]:
        yield self.cisco_style_text()
        for child in sorted(self.children):
            yield from child.lines(sectional_exiting=sectional_exiting)

        if sectional_exiting and (exit_text := self.sectional_exit):
            depth = (
                self.depth() - 1
                if self.sectional_exit_text_parent_level
                else self.depth()
            )
            yield " " * self.driver.rules.indentation * depth + exit_text

    @property
    def sectional_exit(self) -> str | None:
        for rule in self.driver.rules.sectional_exiting:
            if self.is_lineage_match(rule.match_rules):
                if exit_text := rule.exit_text:
                    return exit_text
                return None

        if not self.children:
            return None

        return "exit"

    @property
    def sectional_exit_text_parent_level(self) -> bool:
        for rule in self.driver.rules.sectional_exiting:
            if self.is_lineage_match(rule.match_rules):
                return rule.exit_text_parent_level

        return False

    def delete_sectional_exit(self) -> None:
        try:
            potential_exit = self.children[-1]
        except IndexError:
            return

        if (exit_text := self.sectional_exit) and exit_text == potential_exit.text:
            potential_exit.delete()

    def depth(self) -> int:
        """Returns the distance to the root HConfig object i.e. indent level."""
        return self.parent.depth() + 1

    def move(self, new_parent: HConfig | HConfigChild) -> None:
        """Move one HConfigChild object to different HConfig parent object.

        .. code:: python

            hier1 = config_for_platform(host.platform)
            interface1 = hier1.add_child('interface Vlan2')
            interface1.add_child('ip address 10.0.0.1 255.255.255.252')

            hier2 = Hconfig(host)

            interface1.move(hier2)

        :param new_parent: HConfigChild object -> type list
        """
        new_parent.children.append(self)
        self.delete()

    def lineage(self) -> Iterator[HConfigChild]:
        """Yields the lineage of parent objects up to, but excluding, the root."""
        yield from self.parent.lineage()
        yield self

    def path(self) -> Iterator[str]:
        """Yields the text attribute of child objects up to, but excluding, the root."""
        for child in self.lineage():
            yield child.text

    def cisco_style_text(
        self,
        style: str = "without_comments",
        tag: str | None = None,
    ) -> str:
        """Return a Cisco style formated line i.e. indentation_level + text ! comments."""
        comments: list[str] = []
        if style == "without_comments":
            pass
        elif style == "merged":
            # count the number of instances that have the tag
            instance_count = 0
            instance_comments: set[str] = set()
            for instance in self.instances:
                if tag is None or tag in instance.tags:
                    instance_count += 1
                    instance_comments.update(instance.comments)

            # should the word 'instance' be plural?
            word = "instance" if instance_count == 1 else "instances"

            comments.append(f"{instance_count} {word}")
            comments.extend(instance_comments)
        elif style == "with_comments":
            comments.extend(self.comments)

        comments_str = f" !{', '.join(sorted(comments))}" if comments else ""
        return f"{self.indentation}{self.text}{comments_str}"

    @property
    def indentation(self) -> str:
        return " " * self.driver.rules.indentation * (self.depth() - 1)

    def delete(self) -> None:
        """Delete the current object from its parent."""
        self.parent.children.delete(self)

    def tags_add(self, tag: str | Iterable[str]) -> None:
        """Add a tag to self._tags on all leaf nodes."""
        if self.is_branch:
            for child in self.children:
                child.tags_add(tag)
        elif isinstance(tag, str):
            self._tags.add(tag)
        else:
            self._tags.update(tag)

    def tags_remove(self, tag: str | Iterable[str]) -> None:
        """Remove a tag from self._tags on all leaf nodes."""
        if self.is_branch:
            for child in self.children:
                child.tags_remove(tag)
        elif isinstance(tag, str):
            self._tags.remove(tag)
        else:
            self._tags.difference_update(tag)

    def negate(self) -> HConfigChild:
        """Negate self.text using driver-specific negation rules.

        Negation is resolved in the following priority order:

        1. ``negate_with`` rule — replaces ``self.text`` with a custom
           negation string defined in the driver (e.g. ``no ip route``).
        2. ``negation_default_when`` rule — rewrites the command to its
           ``default`` form (e.g. ``no shutdown`` → ``default shutdown``).
        3. ``negation_sub`` rule — applies a regex substitution to the
           negated text (e.g. truncating after a specific token).
        4. ``swap_negation`` — toggles the negation prefix/declaration
           prefix (e.g. ``shutdown`` ↔ ``no shutdown``).

        Returns self so that callers can chain further operations.
        """
        if negate_with := self.driver.negate_with(self):
            self.text = negate_with
            return self

        if self.use_default_for_negation(self):
            return self._default()

        for rule in self.driver.rules.negation_sub:
            if self.is_lineage_match(rule.match_rules):
                self.text = sub(
                    rule.search,
                    rule.replace,
                    f"{self.driver.negation_prefix}{self.text}",
                )
                return self

        return self.driver.swap_negation(self)

    def use_default_for_negation(self, config: HConfigChild) -> bool:
        return any(
            config.is_lineage_match(rule.match_rules)
            for rule in self.driver.rules.negation_default_when
        )

    @property
    def is_leaf(self) -> bool:
        """Returns True if there are no children and is not an instance of HConfig."""
        return not self.is_branch

    @property
    def is_branch(self) -> bool:
        """Returns True if there are children or is an instance of HConfig."""
        return bool(self.children)

    @property
    def tags(self) -> frozenset[str]:
        """Recursive access to tags on all leaf nodes."""
        if self.is_branch:
            found_tags: set[str] = set()
            for child in self.children:
                found_tags.update(child.tags)
            return frozenset(found_tags)

        return frozenset(self._tags)

    @tags.setter
    def tags(self, value: frozenset[str]) -> None:
        """Recursive access to tags on all leaf nodes."""
        if self.is_branch:
            for child in self.children:
                child.tags = value
        else:
            self._tags = set(value)

    def is_idempotent_command(self, other_children: Iterable[HConfigChild]) -> bool:
        """Determine if self.text is an idempotent change."""
        # Avoid list commands from matching as idempotent
        for rule in self.driver.rules.idempotent_commands_avoid:
            if self.is_lineage_match(rule.match_rules):
                return False

        # Idempotent command identification
        return bool(self.driver.idempotent_for(self, other_children))

    def use_sectional_overwrite_without_negation(self) -> bool:
        """Check self's text to see if negation should be handled by
        overwriting the section without first negating it.
        """
        return any(
            self.is_lineage_match(rule.match_rules)
            for rule in self.driver.rules.sectional_overwrite_no_negate
        )

    def use_sectional_overwrite(self) -> bool:
        """Determines if self.text matches a sectional overwrite rule."""
        return any(
            self.is_lineage_match(rule.match_rules)
            for rule in self.driver.rules.sectional_overwrite
        )

    def overwrite_with(
        self,
        target: HConfigChild,
        delta: HConfig | HConfigChild,
        *,
        negate: bool = True,
    ) -> None:
        """Overwrite self's section in delta with a deep copy of target.

        When the children of self and target differ, this method mutates
        ``delta`` in-place: the existing entry for ``self.text`` is negated
        (if ``negate=True``) or simply deleted (if ``negate=False``), and a
        fresh deep copy of ``target`` is appended.  A ``"re-create section"``
        comment is attached to the new entry, and a ``"dropping section"``
        comment is added to the negated entry when applicable.

        Used by :meth:`_config_to_get_to_right` when a sectional-overwrite
        rule is active for ``self.text``.
        """
        if self.children != target.children:
            if negate:
                if negated := delta.children.get(self.text):
                    negated.negate()
                else:
                    negated = delta.add_child(
                        self.text, check_if_present=False
                    ).negate()

                negated.comments.add("dropping section")
            else:
                delta.children.delete(self.text)
            if self.children:
                new_item = delta.add_deep_copy_of(target)
                new_item.comments.add("re-create section")

    def line_inclusion_test(
        self,
        include_tags: Iterable[str],
        exclude_tags: Iterable[str],
    ) -> bool:
        """Given the line_tags, include_tags, and exclude_tags,
        determine if the line should be included.
        """
        include_line = False

        if include_tags:
            include_line = bool(self.tags.intersection(include_tags))
        if exclude_tags and (include_line or not include_tags):
            return not bool(self.tags.intersection(exclude_tags))

        return include_line

    @property
    def instance(self) -> Instance:
        return Instance(
            id=id(self.root),
            comments=frozenset(self.comments),
            tags=frozenset(self.tags),
        )

    def all_children_sorted_by_tags(
        self,
        include_tags: Iterable[str],
        exclude_tags: Iterable[str],
    ) -> Iterator[HConfigChild]:
        """Yield all children recursively that match include/exclude tags."""
        if self.is_leaf:
            if self.line_inclusion_test(include_tags, exclude_tags):
                yield self
        else:
            self_iter = iter((self,))
            for child in sorted(self.children):
                included_children = child.all_children_sorted_by_tags(
                    include_tags,
                    exclude_tags,
                )
                if peek := next(included_children, None):
                    yield from chain(self_iter, (peek,), included_children)

    def is_lineage_match(self, rules: tuple[MatchRule, ...]) -> bool:
        """A generic test against a lineage of HConfigChild objects."""
        lineage = tuple(self.lineage())

        return len(rules) == len(lineage) and all(
            child.is_match(
                equals=rule.equals,
                startswith=rule.startswith,
                endswith=rule.endswith,
                contains=rule.contains,
                re_search=rule.re_search,
            )
            for (child, rule) in zip(reversed(lineage), reversed(rules), strict=True)
        )

    def is_match(  # noqa: PLR0911
        self,
        *,
        equals: str | SetLikeOfStr | None = None,
        startswith: str | tuple[str, ...] | None = None,
        endswith: str | tuple[str, ...] | None = None,
        contains: str | tuple[str, ...] | None = None,
        re_search: str | None = None,
    ) -> bool:
        """Return True if ``self.text`` satisfies all supplied criteria.

        All arguments are optional.  When *all* arguments are ``None`` the
        method returns ``True`` (matches everything).  When multiple arguments
        are provided, **all** must match.

        Args:
            equals: Exact string match, or a frozenset of acceptable values.
            startswith: ``str.startswith`` prefix (str or tuple of strs).
            endswith: ``str.endswith`` suffix (str or tuple of strs).
            contains: Substring(s) that must appear in ``self.text``.
            re_search: Regular expression applied via :func:`re.search`.

        Returns:
            ``True`` if every non-``None`` criterion is satisfied.

        """
        # Equals filter
        if isinstance(equals, str):
            if self.text != equals:
                return False
        elif (  # pylint: disable=confusing-consecutive-elif
            isinstance(equals, frozenset) and self.text not in equals
        ):
            return False

        # Startswith filter
        if isinstance(startswith, (str, tuple)) and not self.text.startswith(
            startswith
        ):
            return False

        # Regex filter
        if isinstance(re_search, str) and not search(re_search, self.text):
            return False

        # The below filters are less commonly used
        # Endswith filter
        if isinstance(endswith, (str, tuple)) and not self.text.endswith(endswith):
            return False

        # Contains filter
        if isinstance(contains, str):
            if contains not in self.text:
                return False
        elif isinstance(  # pylint: disable=confusing-consecutive-elif
            contains,
            tuple,
        ) and not any(c in self.text for c in contains):
            return False

        return True

    def add_children_deep(self, lines: Iterable[str]) -> HConfigChild:
        """Add child instances of HConfigChild deeply."""
        base = self
        for line in lines:
            base = base.add_child(line)
        return base

    def _default(self) -> HConfigChild:
        """Default self.text."""
        self.text = f"default {self.text_without_negation}"
        return self

    def instantiate_child(self, text: str) -> HConfigChild:
        return HConfigChild(self, text)

    def _is_duplicate_child_allowed(self) -> bool:
        """Determine if duplicate(identical text) children are allowed under the parent."""
        return any(
            self.is_lineage_match(rule.match_rules)
            for rule in self.driver.rules.parent_allows_duplicate_child
        )



from collections.abc import Iterable
from ipaddress import AddressValueError, IPv4Address, IPv4Interface
from re import sub

from hier_config.child import HConfigChild
from hier_config.platforms.models import (
    InterfaceDot1qMode,
    InterfaceDuplex,
    NACHostMode,
    StackMember,
    Vlan,
)
from hier_config.platforms.view_base import (
    ConfigViewInterfaceBase,
    HConfigViewBase,
)


class ConfigViewInterfaceCiscoIOSXR(ConfigViewInterfaceBase):  # noqa: PLR0904
    """Interface config view for Cisco IOS XR."""

    @property
    def _bundle_prefix(self) -> str:
        return "Bundle-Ether"

    @property
    def bundle_id(self) -> str | None:
        raise NotImplementedError

    @property
    def bundle_member_interfaces(self) -> Iterable[str]:
        raise NotImplementedError

    @property
    def bundle_name(self) -> str | None:
        if self.bundle_id:
            return f"{self._bundle_prefix}{self.bundle_id}"
        return None

    @property
    def description(self) -> str:
        if child := self.config.get_child(startswith="description "):
            return child.text.split(maxsplit=1)[1]
        return ""

    @property
    def duplex(self) -> InterfaceDuplex:
        raise NotImplementedError

    @property
    def enabled(self) -> bool:
        raise NotImplementedError

    @property
    def has_nac(self) -> bool:
        """Determine if the interface has NAC configured."""
        raise NotImplementedError

    @property
    def ipv4_interface(self) -> IPv4Interface | None:
        return next(iter(self.ipv4_interfaces), None)

    @property
    def ipv4_interfaces(self) -> Iterable[IPv4Interface]:
        for ipv4_address_obj in self.config.get_children(startswith="ipv4 address "):
            ipv4_address = ipv4_address_obj.text.split()
            try:
                yield IPv4Interface("/".join(ipv4_address[2:4]))
            except AddressValueError:
                continue

    @property
    def is_bundle(self) -> bool:
        return self.name.lower().startswith(self._bundle_prefix)

    @property
    def is_loopback(self) -> bool:
        return self.name.lower().startswith("loopback")

    @property
    def is_subinterface(self) -> bool:
        return "." in self.name

    @property
    def is_svi(self) -> bool:
        return self.name.lower().startswith("vlan")

    @property
    def module_number(self) -> int | None:
        words = self.number.split("/", 1)
        if len(words) == 1:
            return None
        return int(words[0])

    @property
    def nac_control_direction_in(self) -> bool:
        """Determine if the interface has NAC control direction in configured."""
        raise NotImplementedError

    @property
    def nac_host_mode(self) -> NACHostMode | None:
        """Determine the NAC host mode."""
        raise NotImplementedError

    @property
    def nac_mab_first(self) -> bool:
        """Determine if the interface has NAC configured for MAB first."""
        raise NotImplementedError

    @property
    def nac_max_dot1x_clients(self) -> int:
        """Determine the max dot1x clients."""
        raise NotImplementedError

    @property
    def nac_max_mab_clients(self) -> int:
        """Determine the max mab clients."""
        raise NotImplementedError

    @property
    def name(self) -> str:
        return self.config.text.split()[1]

    @property
    def native_vlan(self) -> int | None:
        raise NotImplementedError

    @property
    def number(self) -> str:
        return sub(r"^[a-zA-Z-]+", "", self.name)

    @property
    def parent_name(self) -> str | None:
        if self.is_subinterface:
            return self.name.split(".")[0]
        return None

    @property
    def poe(self) -> bool:
        raise NotImplementedError

    @property
    def port_number(self) -> int:
        return int(self.name.split("/")[-1].split(".")[0])

    @property
    def speed(self) -> tuple[int, ...] | None:
        raise NotImplementedError

    @property
    def subinterface_number(self) -> int | None:
        return int(self.name.split(".")[0 - 1]) if self.is_subinterface else None

    @property
    def tagged_all(self) -> bool:
        raise NotImplementedError

    @property
    def tagged_vlans(self) -> tuple[int, ...]:
        raise NotImplementedError

    @property
    def vrf(self) -> str:
        raise NotImplementedError


class HConfigViewCiscoIOSXR(HConfigViewBase):
    """Full-tree config view for Cisco IOS XR."""

    def dot1q_mode_from_vlans(
        self,
        untagged_vlan: int | None = None,
        tagged_vlans: tuple[int, ...] = (),
        *,
        tagged_all: bool = False,
    ) -> InterfaceDot1qMode | None:
        raise NotImplementedError

    @property
    def hostname(self) -> str | None:
        if child := self.config.get_child(startswith="hostname "):
            return child.text.split()[1].lower()
        return None

    @property
    def interface_names_mentioned(self) -> frozenset[str]:
        raise NotImplementedError

    @property
    def interface_views(self) -> Iterable[ConfigViewInterfaceCiscoIOSXR]:
        for interface in self.interfaces:
            yield ConfigViewInterfaceCiscoIOSXR(interface)

    @property
    def interfaces(self) -> Iterable[HConfigChild]:
        return self.config.get_children(startswith="interface ")

    @property
    def ipv4_default_gw(self) -> IPv4Address | None:
        raise NotImplementedError

    @property
    def location(self) -> str:
        raise NotImplementedError

    @property
    def stack_members(self) -> Iterable[StackMember]:
        raise NotImplementedError

    @property
    def vlans(self) -> Iterable[Vlan]:
        raise NotImplementedError



from collections.abc import Callable, Iterable
from pathlib import Path
from typing import Any

import yaml
from pydantic import TypeAdapter

from hier_config import Platform, get_hconfig_driver
from hier_config.models import (
    FullTextSubRule,
    IdempotentCommandsAvoidRule,
    IdempotentCommandsRule,
    IndentAdjustRule,
    MatchRule,
    NegationDefaultWhenRule,
    NegationDefaultWithRule,
    NegationSubRule,
    OrderingRule,
    ParentAllowsDuplicateChildRule,
    PerLineSubRule,
    ReferenceLocation,
    SectionalExitingRule,
    SectionalOverwriteNoNegateRule,
    SectionalOverwriteRule,
    TagRule,
    UnusedObjectRule,
)
from hier_config.platforms.driver_base import HConfigDriverBase

HCONFIG_PLATFORM_V2_TO_V3_MAPPING = {
    "ios": Platform.CISCO_IOS,
    "iosxe": Platform.CISCO_IOS,
    "iosxr": Platform.CISCO_XR,
    "nxos": Platform.CISCO_NXOS,
    "eos": Platform.ARISTA_EOS,
    "junos": Platform.JUNIPER_JUNOS,
    "vyos": Platform.VYOS,
}


def _set_match_rule(lineage: dict[str, Any]) -> MatchRule | None:
    if startswith := lineage.get("startswith"):
        return MatchRule(startswith=startswith)
    if endswith := lineage.get("endswith"):
        return MatchRule(endswith=endswith)
    if contains := lineage.get("contains"):
        return MatchRule(contains=contains)
    if equals := lineage.get("equals"):
        return MatchRule(equals=equals)
    if re_search := lineage.get("re_search"):
        return MatchRule(re_search=re_search)

    return None


def _collect_match_rules(
    lineages: Iterable[dict[str, Any]],
) -> tuple[MatchRule, ...]:
    collected: list[MatchRule] = []
    for lineage in lineages:
        match_rule = _set_match_rule(lineage)
        if match_rule is not None:
            collected.append(match_rule)
    return tuple(collected)


def read_text_from_file(file_path: str) -> str:
    """Function that loads the contents of a file into memory.

    Args:
        file_path (str): The path to the configuration file.

    Returns:
        str: The configuration file contents as a string.

    """
    return Path(file_path).read_text(encoding="utf-8")


def load_hier_config_tags(tags_file: str) -> tuple[TagRule, ...]:
    """Loads and validates Hier Config tags from a YAML file.

    Args:
        tags_file (str): Path to the YAML file containing the tags.

    Returns:
        Tuple[TagRule, ...]: A tuple of validated TagRule objects.

    """
    tags_data = yaml.safe_load(read_text_from_file(file_path=tags_file))
    return TypeAdapter(tuple[TagRule, ...]).validate_python(tags_data)


def hconfig_v2_os_v3_platform_mapper(os_name: str) -> Platform:
    """Map a Hier Config v2 operating system name to a v3 Platform enumeration.

    Args:
        os_name (str): The name of the OS as defined in Hier Config v2.

    Returns:
        Platform: The corresponding Platform enumeration for Hier Config v3.

    Example:
        >>> hconfig_v2_os_v3_platform_mapper("CISCO_IOS")
        <Platform.CISCO_IOS: 'ios'>

    """
    return HCONFIG_PLATFORM_V2_TO_V3_MAPPING.get(os_name, Platform.GENERIC)


def hconfig_v3_platform_v2_os_mapper(platform: Platform) -> str:
    """Map a Hier Config v3 Platform enumeration to a v2 operating system name.

    Args:
        platform (Platform): A Platform enumeration from Hier Config v3.

    Returns:
        str: The corresponding OS name for Hier Config v2.

    Example:
        >>> hconfig_v3_platform_v2_os_mapper(Platform.CISCO_IOS)
        "ios"

    """
    for os_name, plat in HCONFIG_PLATFORM_V2_TO_V3_MAPPING.items():
        if plat == platform:
            return os_name

    return "generic"


def _process_simple_rules(
    v2_options: dict[str, Any],
    key: str,
    rule_class: type[Any],
    append_to: Callable[[Any], None],
) -> None:
    """Process v2 rules that only need match_rules."""
    for rule in v2_options.get(key, ()):
        match_rules = _collect_match_rules(rule.get("lineage", []))
        append_to(rule_class(match_rules=match_rules))


def _process_custom_rules(
    v2_options: dict[str, Any], driver: HConfigDriverBase
) -> None:
    """Process v2 rules that require custom handling."""
    for rule in v2_options.get("ordering", ()):
        match_rules = _collect_match_rules(rule.get("lineage", []))
        weight = rule.get("order", 500) - 500
        driver.rules.ordering.append(
            OrderingRule(match_rules=match_rules, weight=weight),
        )

    for rule in v2_options.get("indent_adjust", ()):
        driver.rules.indent_adjust.append(
            IndentAdjustRule(
                start_expression=rule.get("start_expression"),
                end_expression=rule.get("end_expression"),
            )
        )

    for rule in v2_options.get("sectional_exiting", ()):
        match_rules = _collect_match_rules(rule.get("lineage", []))
        driver.rules.sectional_exiting.append(
            SectionalExitingRule(
                match_rules=match_rules, exit_text=rule.get("exit_text", "")
            ),
        )

    for rule in v2_options.get("full_text_sub", ()):
        driver.rules.full_text_sub.append(
            FullTextSubRule(
                search=rule.get("search", ""), replace=rule.get("replace", "")
            )
        )

    for rule in v2_options.get("per_line_sub", ()):
        driver.rules.per_line_sub.append(
            PerLineSubRule(
                search=rule.get("search", ""), replace=rule.get("replace", "")
            )
        )

    for rule in v2_options.get("negation_negate_with", ()):
        match_rules = _collect_match_rules(rule.get("lineage", []))
        driver.rules.negate_with.append(
            NegationDefaultWithRule(match_rules=match_rules, use=rule.get("use", "")),
        )

    for rule in v2_options.get("negation_sub", ()):
        match_rules = _collect_match_rules(rule.get("lineage", []))
        driver.rules.negation_sub.append(
            NegationSubRule(
                match_rules=match_rules,
                search=rule.get("search", ""),
                replace=rule.get("replace", ""),
            ),
        )

    for rule in v2_options.get("unused_objects", ()):
        match_rules = _collect_match_rules(rule.get("lineage", []))
        ref_locations = tuple(
            ReferenceLocation(
                match_rules=_collect_match_rules(ref.get("lineage", [])),
                reference_re=ref.get("reference_re", ""),
            )
            for ref in rule.get("reference_locations", [])
        )
        driver.rules.unused_objects.append(
            UnusedObjectRule(
                match_rules=match_rules,
                name_re=rule.get("name_re", ""),
                reference_locations=ref_locations,
            ),
        )


def load_hconfig_v2_options(
    v2_options: dict[str, Any] | str, platform: Platform
) -> HConfigDriverBase:
    """Load Hier Config v2 options to v3 driver format from either a dictionary or a file.

    Args:
        v2_options (Union[dict, str]): Either a dictionary containing v2 options or
            a file path to a YAML file containing the v2 options.
        platform (Platform): The Hier Config v3 Platform enum for the target platform.

    Returns:
        HConfigDriverBase: A v3 driver instance with the migrated rules.

    """
    if isinstance(v2_options, str):
        v2_options = yaml.safe_load(read_text_from_file(file_path=v2_options))

    if not isinstance(v2_options, dict):
        msg = "v2_options must be a dictionary or a valid file path."
        raise TypeError(msg)

    driver = get_hconfig_driver(platform)

    # Process simple rules that only need match_rules
    simple_rules: tuple[tuple[str, type[Any], Callable[[Any], None]], ...] = (
        (
            "sectional_overwrite",
            SectionalOverwriteRule,
            driver.rules.sectional_overwrite.append,
        ),
        (
            "sectional_overwrite_no_negate",
            SectionalOverwriteNoNegateRule,
            driver.rules.sectional_overwrite_no_negate.append,
        ),
        (
            "parent_allows_duplicate_child",
            ParentAllowsDuplicateChildRule,
            driver.rules.parent_allows_duplicate_child.append,
        ),
        (
            "idempotent_commands_blacklist",
            IdempotentCommandsAvoidRule,
            driver.rules.idempotent_commands_avoid.append,
        ),
        (
            "idempotent_commands",
            IdempotentCommandsRule,
            driver.rules.idempotent_commands.append,
        ),
        (
            "negation_default_when",
            NegationDefaultWhenRule,
            driver.rules.negation_default_when.append,
        ),
    )
    for key, rule_class, append_to in simple_rules:
        _process_simple_rules(v2_options, key, rule_class, append_to)

    # Process rules that require custom handling
    _process_custom_rules(v2_options, driver)

    return driver


def load_hconfig_v2_options_from_file(
    options_file: str, platform: Platform
) -> HConfigDriverBase:
    """Load Hier Config v2 options file to v3 driver format.

    Args:
        options_file (str): The v2 options file.
        platform (Platform): The Hier Config v3 Platform enum for the target platform.

    Returns:
        HConfigDriverBase: A v3 driver instance with the migrated rules.

    """
    hconfig_options = yaml.safe_load(read_text_from_file(file_path=options_file))
    return load_hconfig_v2_options(v2_options=hconfig_options, platform=platform)


def load_hconfig_v2_tags(
    v2_tags: list[dict[str, Any]] | str,
) -> tuple["TagRule"] | tuple["TagRule", ...]:
    """Convert v2-style tags into v3-style TagRule Pydantic objects for Hier Config.

    Args:
        v2_tags (Union[list[dict[str, Any]], str]):
            Either a list of dictionaries representing v2-style tags or a file path
            to a YAML file containing the v2-style tags.
            - If a list is provided, each dictionary should contain:
              - `lineage`: A list of dictionaries with rules (e.g., `startswith`, `endswith`).
              - `add_tags`: A string representing the tag to add.
            - If a file path is provided, it will be read and parsed as YAML.

    Returns:
        Tuple[TagRule]: A tuple of TagRule Pydantic objects representing v3-style tags.

    """
    # Load tags from a file if a string is provided
    if isinstance(v2_tags, str):
        v2_tags = yaml.safe_load(read_text_from_file(file_path=v2_tags))

    # Ensure v2_tags is a list
    if not isinstance(v2_tags, list):
        msg = "v2_tags must be a list of dictionaries or a valid file path."
        raise TypeError(msg)

    v3_tags: list[TagRule] = []

    for v2_tag in v2_tags:
        if "lineage" in v2_tag and "add_tags" in v2_tag:
            # Extract the v2 fields
            lineage_rules = v2_tag["lineage"]
            tags = v2_tag["add_tags"]

            # Convert to MatchRule objects
            match_rules = tuple(
                match_rule
                for lineage in lineage_rules
                if (match_rule := _set_match_rule(lineage)) is not None
            )

            # Create the TagRule object
            v3_tag = TagRule(match_rules=match_rules, apply_tags=frozenset([tags]))
            v3_tags.append(v3_tag)

    return tuple(v3_tags)



from __future__ import annotations

from typing import TYPE_CHECKING, TypeVar, overload

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator

    from hier_config import HConfigChild

_D = TypeVar("_D")


class HConfigChildren:
    """Ordered collection of `HConfigChild` objects with fast text-keyed look-up.

    Internally maintains both a `list` (for ordered iteration) and a `dict`
    (for O(1) membership and retrieval by `child.text`).  When duplicate child
    text is allowed by the driver, the mapping always points to the *first*
    occurrence while the list preserves all entries in insertion order.
    """

    def __init__(self) -> None:
        self._data: list[HConfigChild] = []
        self._mapping: dict[str, HConfigChild] = {}

    @overload
    def __getitem__(self, subscript: int | str) -> HConfigChild: ...

    @overload
    def __getitem__(self, subscript: slice) -> list[HConfigChild]: ...

    def __getitem__(
        self, subscript: slice | int | str
    ) -> HConfigChild | list[HConfigChild]:
        if isinstance(subscript, slice):
            return self._data[subscript]
        if isinstance(subscript, int):
            return self._data[subscript]
        return self._mapping[subscript]

    def __setitem__(self, index: int, child: HConfigChild) -> None:
        self._data[index] = child
        self.rebuild_mapping()

    def __contains__(self, item: str) -> bool:
        return item in self._mapping

    def __iter__(self) -> Iterator[HConfigChild]:
        return iter(self._data)

    def __len__(self) -> int:
        return len(self._data)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, HConfigChildren):
            return NotImplemented

        self_len = len(self._data)
        other_len = len(other._data)
        # Superfast succeed method for no children
        if self_len == other_len == 0:
            return True

        # Superfast fail method
        if self_len != other_len:
            return False

        # Fast fail method
        if self._mapping.keys() != other._mapping.keys():
            return False

        # Slower full comparison
        return all(
            self_child == other_child
            for self_child, other_child in zip(
                sorted(self._data),
                sorted(other._data),
                strict=False,
            )
        )

    def __hash__(self) -> int:
        return hash(
            (*self._data,),
        )

    def append(
        self,
        child: HConfigChild,
        *,
        update_mapping: bool = True,
    ) -> HConfigChild:
        self._data.append(child)
        if update_mapping:
            self._mapping.setdefault(child.text, child)

        return child

    def clear(self) -> None:
        """Delete all children."""
        self._data.clear()
        self._mapping.clear()

    def delete(self, child_or_text: HConfigChild | str) -> None:
        """Delete a child from self._data and self._mapping."""
        if isinstance(child_or_text, str):
            if child_or_text in self._mapping:
                self._data[:] = [c for c in self._data if c.text != child_or_text]
                self.rebuild_mapping()
        else:
            old_len = len(self._data)
            self._data = [c for c in self._data if c is not child_or_text]
            if old_len != len(self._data):
                self.rebuild_mapping()

    def extend(self, children: Iterable[HConfigChild]) -> None:
        """Add child instances of HConfigChild."""
        self._data.extend(children)
        for child in children:
            self._mapping.setdefault(child.text, child)

    def get(self, key: str, default: _D | None = None) -> HConfigChild | _D | None:
        return self._mapping.get(key, default)

    def index(self, child: HConfigChild) -> int:
        return self._data.index(child)

    def rebuild_mapping(self) -> None:
        """Rebuild self._mapping."""
        self._mapping.clear()
        for child in self._data:
            self._mapping.setdefault(child.text, child)



from hier_config.child import HConfigChild
from hier_config.platforms.driver_base import HConfigDriverBase, HConfigDriverRules
from hier_config.platforms.functions import convert_to_set_commands


class HConfigDriverJuniperJUNOS(HConfigDriverBase):  # pylint: disable=too-many-instance-attributes
    """Driver for Juniper JunOS (experimental).

    Converts hierarchical JunOS configuration into flat ``set``/``delete``
    command syntax via a preprocessor.  Overrides ``declaration_prefix`` to
    ``"set "`` and ``negation_prefix`` to ``"delete "`` so that
    ``swap_negation`` correctly toggles between the two forms.

    .. warning::
        JunOS support is experimental and has not been extensively tested in
        production environments.  Use with caution.

    Platform enum: ``Platform.JUNIPER_JUNOS``.
    """

    def swap_negation(self, child: HConfigChild) -> HConfigChild:
        """Swap negation of a `self.text`."""
        if child.text.startswith(self.negation_prefix):
            child.text = f"{self.declaration_prefix}{child.text_without_negation}"
        elif child.text.startswith(self.declaration_prefix):
            child.text = f"{self.negation_prefix}{child.text.removeprefix(self.declaration_prefix)}"
        else:
            message = f"{child.text=} did not start with {self.negation_prefix} or {self.declaration_prefix}."
            raise ValueError(message)

        return child

    @property
    def declaration_prefix(self) -> str:
        return "set "

    @property
    def negation_prefix(self) -> str:
        return "delete "

    @staticmethod
    def config_preprocessor(config_text: str) -> str:
        return convert_to_set_commands(config_text)

    @staticmethod
    def _instantiate_rules() -> HConfigDriverRules:
        return HConfigDriverRules()



def expand_range(number_range_str: str) -> tuple[int, ...]:
    """Expand ranges like 2-5,8,22-45."""
    numbers: list[int] = []
    for number_range in number_range_str.split(","):
        start_stop = number_range.split("-")
        if len(start_stop) == 2:
            start = int(start_stop[0])
            stop = int(start_stop[1])
            numbers.extend(n for n in range(start, stop + 1))
        else:
            numbers.append(int(start_stop[0]))
    if len(set(numbers)) != len(numbers):
        message = "len(set(numbers)) must be equal to len(numbers)."
        raise ValueError(message)
    return tuple(numbers)


def convert_to_set_commands(config_raw: str) -> str:
    """Convert a Juniper style config string into a list of set commands.

    Args:
        config_raw (str): The config string to convert to set commands
    Returns:
        config_raw (str): Configuration string

    """
    lines = config_raw.split("\n")
    path: list[str] = []
    set_commands: list[str] = []

    for line in lines:
        stripped_line = line.strip()

        # Skip empty lines
        if not stripped_line:
            continue

        # Strip ; from the end of the line
        stripped_line = stripped_line.removesuffix(";")

        # Count the number of spaces at the beginning to determine the level
        level = line.find(stripped_line) // 4

        # Adjust the current path based on the level
        path = path[:level]

        # If the line ends with '{' or '}', it starts a new block
        if stripped_line.endswith(("{", "}")):
            path.append(stripped_line[:-1].strip())
        elif stripped_line.startswith(("set", "delete")):
            # It's already a set command, so just add it to the list
            set_commands.append(stripped_line)
        else:
            # It's a command line, construct the full command
            command = f"set {' '.join(path)} {stripped_line}"
            set_commands.append(command)

    return "\n".join(set_commands)



import re
from collections.abc import Iterable
from ipaddress import AddressValueError, IPv4Address, IPv4Interface

from hier_config.child import HConfigChild
from hier_config.platforms.functions import expand_range
from hier_config.platforms.hp_procurve.functions import hp_procurve_expand_range
from hier_config.platforms.models import (
    InterfaceDot1qMode,
    InterfaceDuplex,
    NACHostMode,
    StackMember,
    Vlan,
)
from hier_config.platforms.view_base import (
    ConfigViewInterfaceBase,
    HConfigViewBase,
)


class ConfigViewInterfaceHPProcurve(  # noqa: PLR0904 pylint: disable=abstract-method
    ConfigViewInterfaceBase,
):
    """Interface config view for HP ProCurve / Aruba AOSS."""

    @property
    def bundle_id(self) -> str | None:
        raise NotImplementedError

    @property
    def bundle_member_interfaces(self) -> Iterable[str]:
        # trunk 1/45,2/45 trk1 trunk
        bundle = self.config.parent.get_child(
            re_search=rf"^trunk .* {self.name.lower()} (trunk|lacp)$",
        )
        if self.is_bundle and bundle is None:
            message = (
                f"Interface is a bundle but bundle config was not found: {self.name}"
            )
            raise TypeError(message)
        if bundle is None:
            message = f"The bundle config line couldn't be found: {self.name}"
            raise ValueError(message)
        return hp_procurve_expand_range(bundle.text.split()[1])

    @property
    def bundle_name(self) -> str | None:
        for bundle_def in self.config.parent.get_children(startswith="trunk "):
            interface_range = bundle_def.text.split()[1]
            interfaces = hp_procurve_expand_range(interface_range)
            if self.name in interfaces:
                # Capitalizing the interface name is consistent with 1/A1 interface naming
                # and is consistent with references under `vlan 10/n  tagged Trk1`
                return bundle_def.text.split()[2].capitalize()
        return None

    @property
    def description(self) -> str:
        if child := self.config.get_child(startswith="name "):
            return child.text.split(maxsplit=1)[1].replace('"', "")
        return ""

    @property
    def duplex(self) -> InterfaceDuplex:
        if duplex := self.config.get_child(startswith="speed-duplex "):
            return _duplex_from_speed_duplex(duplex.text)
        return InterfaceDuplex.AUTO

    @property
    def enabled(self) -> bool:
        return not self.config.get_child(equals="disable")

    @property
    def has_nac(self) -> bool:
        """Determine if the interface has NAC configured."""
        return any(
            line in self.config.parent.children
            for line in (
                f"aaa port-access authenticator {self.name}",
                f"aaa port-access mac-based {self.name}",
            )
        )

    @property
    def ipv4_interface(self) -> IPv4Interface | None:
        return next(iter(self.ipv4_interfaces), None)

    @property
    def ipv4_interfaces(self) -> Iterable[IPv4Interface]:
        for ipv4_address_obj in self.config.get_children(startswith="ip address "):
            ipv4_address = ipv4_address_obj.text.split()
            try:
                yield IPv4Interface("/".join(ipv4_address[2:4]))
            except AddressValueError:
                continue

    @property
    def is_bundle(self) -> bool:
        return self.name.lower().startswith(self._bundle_prefix)

    @property
    def is_loopback(self) -> bool:
        return self.name.lower().startswith("loopback")

    @property
    def is_subinterface(self) -> bool:
        return "." in self.name

    @property
    def is_svi(self) -> bool:
        return self.name.lower().startswith("vlan")

    @property
    def module_number(self) -> int | None:
        words = self.number.split("/", 1)
        if len(words) == 1:
            return None
        return int(words[0])

    @property
    def nac_control_direction_in(self) -> bool:
        """Determine if the interface has NAC control direction in configured."""
        return (
            f"aaa port-access {self.name} controlled-direction in"
            in self.config.parent.children
        )

    @property
    def nac_host_mode(self) -> NACHostMode | None:
        """Determine the NAC host mode."""
        # hp_procurve does not support host mode
        return None

    @property
    def nac_mab_first(self) -> bool:
        """Determine if the interface has NAC configured for MAB first."""
        return bool(
            self.config.parent.get_child(
                equals=f"aaa port-access {self.name} auth-order mac-based authenticator",
            ),
        )

    @property
    def nac_max_dot1x_clients(self) -> int:
        """Determine the max dot1x clients."""
        if child := self.config.parent.get_child(
            startswith=f"aaa port-access authenticator {self.name} client-limit ",
        ):
            return int(child.text.split()[5])
        return 1

    @property
    def nac_max_mab_clients(self) -> int:
        """Determine the max mab clients."""
        if child := self.config.parent.get_child(
            startswith=f"aaa port-access mac-based {self.name} addr-limit ",
        ):
            return int(child.text.split()[5])
        return 1

    @property
    def name(self) -> str:
        if self.config.text.startswith("interface "):
            return self.config.text.split()[1]
        return self.config.text

    @property
    def native_vlan(self) -> int | None:
        if vlan := self.config.get_child(startswith="untagged vlan "):
            return int(vlan.text.split()[2])
        return None

    @property
    def number(self) -> str:
        return re.sub(r"^[a-zA-Z-]+", "", self.name)

    @property
    def parent_name(self) -> str | None:
        if self.is_subinterface:
            return self.name.split(".")[0]
        return None

    @property
    def poe(self) -> bool:
        return not self.config.get_child(equals="no power-over-ethernet")

    @property
    def port_number(self) -> int:
        return int(self.name.split("/")[-1].split(".")[0])

    @property
    def speed(self) -> tuple[int, ...] | None:
        if speed := self.config.get_child(startswith="speed-duplex "):
            return _speed_from_speed_duplex(speed.text)
        return None

    @property
    def subinterface_number(self) -> int | None:
        return int(self.name.split(".")[0 - 1]) if self.is_subinterface else None

    @property
    def tagged_all(self) -> bool:
        return False

    @property
    def tagged_vlans(self) -> tuple[int, ...]:
        return tuple(
            int(c.text.split()[2])
            for c in self.config.get_children(startswith="tagged vlan ")
        )

    @property
    def vrf(self) -> str:
        return ""

    @property
    def _bundle_prefix(self) -> str:
        return "trk"


def _speed_from_speed_duplex(speed_duplex: str) -> tuple[int, ...] | None:
    if speed_duplex.startswith("10"):
        return (int(speed_duplex.split("-", maxsplit=1)[0]),)
    if speed_duplex.startswith("auto-"):
        return tuple(int(s) for s in speed_duplex.replace("auto-", "").split("-"))
    return None


def _duplex_from_speed_duplex(speed_duplex: str) -> InterfaceDuplex:
    if speed_duplex.startswith("auto"):
        return InterfaceDuplex(speed_duplex[:4])
    if speed_duplex.endswith(("half", "full")):
        return InterfaceDuplex(speed_duplex[-4:])
    return InterfaceDuplex(speed_duplex)


class HConfigViewHPProcurve(HConfigViewBase):
    """Full-tree config view for HP ProCurve / Aruba AOSS."""

    def dot1q_mode_from_vlans(
        self,
        untagged_vlan: int | None = None,
        tagged_vlans: tuple[int, ...] = (),
        *,
        tagged_all: bool = False,
    ) -> InterfaceDot1qMode | None:
        raise NotImplementedError

    @property
    def hostname(self) -> str | None:
        if child := self.config.get_child(startswith="hostname "):
            return child.text.split()[1].lower().replace('"', "")
        return None

    @property
    def interface_names_mentioned(self) -> frozenset[str]:
        interfaces = {model.name for model in self.interface_views}

        for child in self.config.children:
            text = child.text_without_negation
            words = text.split()
            if text.startswith("aaa port-access "):
                found = (
                    words[3] if words[2] in {"authenticator", "mac-based"} else words[2]
                )

                if re.search(r"^(\d+|\d+/\d+)$", found):
                    interfaces.add(found)

        return frozenset(interfaces)

    @property
    def interface_views(self) -> Iterable[ConfigViewInterfaceHPProcurve]:
        for interface in self.interfaces:
            yield ConfigViewInterfaceHPProcurve(interface)
        for vlan in self.config.get_children(startswith="vlan "):
            if vlan.get_child(startswith="ip address "):
                yield ConfigViewInterfaceHPProcurve(vlan)

    @property
    def interfaces(self) -> Iterable[HConfigChild]:
        return self.config.get_children(startswith="interface ")

    @property
    def ipv4_default_gw(self) -> IPv4Address | None:
        if gateway := self.config.get_child(startswith="ip default-gateway "):
            return IPv4Address(gateway.text.split()[2])
        return None

    @property
    def location(self) -> str:
        if location := self.config.get_child(startswith="snmp-server location "):
            return location.text.split(maxsplit=2)[2].replace('"', "")
        return ""

    @property
    def stack_members(self) -> Iterable[StackMember]:
        """Stacking
        member 1 type "JL123" mac-address abc123-abc123
        member 1 priority 255
        member 2 type "JL123" mac-address abc123-abc123
        member 2 priority 254
        ...
        """
        stacking = self.config.get_child(equals="stacking")
        if not stacking:
            return

        for member in stacking.get_children(startswith="member"):
            words = member.text.split()
            if words[2] == "type":
                member_id = int(words[1])
                yield StackMember(
                    id=member_id,
                    priority=256 - member_id,
                    mac_address=words[5],
                    model=words[3].replace('"', ""),
                )

    @property
    def vlans(self) -> Iterable[Vlan]:
        yielded_vlans: set[int] = set()

        # Yield explicitly defined VLANs
        for child in self.config.get_children(re_search="^vlan [0-9,-]+$"):
            vlan_name = None
            if name := child.get_child(startswith="name "):
                _, vlan_name = name.text.split(maxsplit=1)
                vlan_name = vlan_name.replace('"', "")
            for vlan_id in expand_range(child.text.split()[1]):
                yielded_vlans.add(vlan_id)
                yield Vlan(
                    id=vlan_id,
                    name=vlan_name or None,
                )

        # Yield any remaining unnamed VLANs mentioned on interfaces
        for interface_view in self.interface_views:
            for vlan_id in interface_view.tagged_vlans:
                if vlan_id not in yielded_vlans:
                    yielded_vlans.add(vlan_id)
                    yield Vlan(
                        id=vlan_id,
                        name=None,
                    )
            if (
                native_vlan := interface_view.native_vlan
            ) and native_vlan not in yielded_vlans:
                yielded_vlans.add(native_vlan)
                yield Vlan(
                    id=native_vlan,
                    name=None,
                )



from enum import Enum, auto

from pydantic import NonNegativeInt, PositiveInt

from hier_config.models import BaseModel


class NACHostMode(str, Enum):
    """802.1X / NAC host-mode options ordered from most to least secure."""

    # Ordered from the most to the least secure
    SINGLE_HOST = "single-host"
    MULTI_DOMAIN = "multi-domain"
    MULTI_AUTH = "multi-auth"
    MULTI_HOST = "multi-host"


class InterfaceDot1qMode(str, Enum):
    """802.1Q encapsulation mode of a switchport interface."""

    ACCESS = auto()
    TAGGED = auto()
    TAGGED_ALL = auto()


class StackMember(BaseModel):
    """Identity and priority record for a single switch-stack member."""

    id: NonNegativeInt
    priority: NonNegativeInt
    mac_address: str | None  # not defined in cisco_ios stacks
    model: str


class InterfaceDuplex(str, Enum):
    """Physical duplex setting of an Ethernet interface."""

    AUTO = auto()
    FULL = auto()
    HALF = auto()


class Vlan(BaseModel):
    """Identifier and optional name for a single VLAN."""

    id: PositiveInt
    name: str | None



from abc import ABC, abstractmethod
from collections.abc import Callable, Iterable
from re import Match, search

from pydantic import Field, PositiveInt

from hier_config.child import HConfigChild
from hier_config.models import (
    BaseModel,
    FullTextSubRule,
    IdempotentCommandsAvoidRule,
    IdempotentCommandsRule,
    IndentAdjustRule,
    MatchRule,
    NegationDefaultWhenRule,
    NegationDefaultWithRule,
    NegationSubRule,
    OrderingRule,
    ParentAllowsDuplicateChildRule,
    PerLineSubRule,
    SectionalExitingRule,
    SectionalOverwriteNoNegateRule,
    SectionalOverwriteRule,
    UnusedObjectRule,
)
from hier_config.root import HConfig


def _full_text_sub_rules_default() -> list[FullTextSubRule]:
    return []


def _idempotent_commands_rules_default() -> list[IdempotentCommandsRule]:
    return []


def _idempotent_commands_avoid_rules_default() -> list[IdempotentCommandsAvoidRule]:
    return []


def _indent_adjust_rules_default() -> list[IndentAdjustRule]:
    return []


def _negation_default_when_rules_default() -> list[NegationDefaultWhenRule]:
    return []


def _negate_with_rules_default() -> list[NegationDefaultWithRule]:
    return []


def _ordering_rules_default() -> list[OrderingRule]:
    return []


def _parent_allows_duplicate_child_rules_default() -> list[
    ParentAllowsDuplicateChildRule
]:
    return []


def _per_line_sub_rules_default() -> list[PerLineSubRule]:
    return []


def _post_load_callbacks_default() -> list[Callable[[HConfig], None]]:
    return []


def _sectional_exiting_rules_default() -> list[SectionalExitingRule]:
    return []


def _sectional_overwrite_rules_default() -> list[SectionalOverwriteRule]:
    return []


def _sectional_overwrite_no_negate_rules_default() -> list[
    SectionalOverwriteNoNegateRule
]:
    return []


def _negation_sub_rules_default() -> list[NegationSubRule]:
    return []


def _unused_object_rules_default() -> list[UnusedObjectRule]:
    return []


class HConfigDriverRules(BaseModel):  # pylint: disable=too-many-instance-attributes
    """Pydantic model holding all rule collections for a platform driver.

    Each field corresponds to one category of driver behaviour (e.g. negation,
    ordering, idempotency).  Instantiated by each driver's ``_instantiate_rules``
    static method and stored on :class:`HConfigDriverBase`.
    """

    full_text_sub: list[FullTextSubRule] = Field(
        default_factory=_full_text_sub_rules_default
    )
    idempotent_commands: list[IdempotentCommandsRule] = Field(
        default_factory=_idempotent_commands_rules_default
    )
    idempotent_commands_avoid: list[IdempotentCommandsAvoidRule] = Field(
        default_factory=_idempotent_commands_avoid_rules_default
    )
    indent_adjust: list[IndentAdjustRule] = Field(
        default_factory=_indent_adjust_rules_default
    )
    indentation: PositiveInt = 2
    negation_default_when: list[NegationDefaultWhenRule] = Field(
        default_factory=_negation_default_when_rules_default
    )
    negate_with: list[NegationDefaultWithRule] = Field(
        default_factory=_negate_with_rules_default
    )
    ordering: list[OrderingRule] = Field(default_factory=_ordering_rules_default)
    parent_allows_duplicate_child: list[ParentAllowsDuplicateChildRule] = Field(
        default_factory=_parent_allows_duplicate_child_rules_default
    )
    per_line_sub: list[PerLineSubRule] = Field(
        default_factory=_per_line_sub_rules_default
    )
    post_load_callbacks: list[Callable[[HConfig], None]] = Field(
        default_factory=_post_load_callbacks_default
    )
    sectional_exiting: list[SectionalExitingRule] = Field(
        default_factory=_sectional_exiting_rules_default
    )
    sectional_overwrite: list[SectionalOverwriteRule] = Field(
        default_factory=_sectional_overwrite_rules_default
    )
    sectional_overwrite_no_negate: list[SectionalOverwriteNoNegateRule] = Field(
        default_factory=_sectional_overwrite_no_negate_rules_default
    )
    negation_sub: list[NegationSubRule] = Field(
        default_factory=_negation_sub_rules_default
    )
    unused_objects: list[UnusedObjectRule] = Field(
        default_factory=_unused_object_rules_default
    )


class HConfigDriverBase(ABC):
    """Defines all hier_config options, rules, and rule checking methods.
    Override methods as needed.
    """

    def __init__(self) -> None:
        self.rules = self._instantiate_rules()

    def idempotent_for(
        self,
        config: HConfigChild,
        other_children: Iterable[HConfigChild],
    ) -> HConfigChild | None:
        for rule in self.rules.idempotent_commands:
            if not config.is_lineage_match(rule.match_rules):
                continue

            config_key = self._idempotency_key(config, rule.match_rules)

            for other_child in other_children:
                if not other_child.is_lineage_match(rule.match_rules):
                    continue

                if self._idempotency_key(other_child, rule.match_rules) == config_key:
                    return other_child

        return None

    def negate_with(self, config: HConfigChild) -> str | None:
        for with_rule in self.rules.negate_with:
            if config.is_lineage_match(with_rule.match_rules):
                return with_rule.use
        return None

    def swap_negation(self, child: HConfigChild) -> HConfigChild:
        """Swap negation of a `child.text`."""
        if child.text.startswith(self.negation_prefix):
            child.text = child.text_without_negation
        else:
            child.text = f"{self.negation_prefix}{child.text}"

        return child

    def _idempotency_key(
        self,
        config: HConfigChild,
        match_rules: tuple[MatchRule, ...],
    ) -> tuple[str, ...]:
        """Build a structural identity for `config` that respects driver rules.

        Args:
            config: The child being evaluated for idempotency.
            match_rules: The match rules describing the lineage signature.

        Returns:
            A tuple of string fragments representing the idempotency key.

        """
        lineage = tuple(config.lineage())
        if len(lineage) != len(match_rules):
            return ()

        components: list[str] = []
        for child, rule in zip(lineage, match_rules, strict=False):
            components.append(self._idempotency_component_key(child, rule))
        return tuple(components)

    def _idempotency_component_key(
        self,
        child: HConfigChild,
        rule: MatchRule,
    ) -> str:
        """Derive the structural key for a single lineage component.

        Args:
            child: The lineage child contributing to the key.
            rule: The rule governing how to match the child.

        Returns:
            A string fragment representing the component key.

        """
        text = child.text
        normalized_text = text.removeprefix(self.negation_prefix)

        parts: list[str] = []
        parts.extend(self._key_from_equals(rule.equals, text))
        parts.extend(self._key_from_prefix(rule.startswith, normalized_text))
        parts.extend(self._key_from_suffix(rule.endswith, normalized_text))
        parts.extend(self._key_from_contains(rule.contains, normalized_text))
        parts.extend(self._key_from_regex(rule.re_search, normalized_text, text))

        if not parts:
            parts.append(f"text|{normalized_text}")

        return ";".join(parts)

    @staticmethod
    def _key_from_equals(equals: str | frozenset[str] | None, text: str) -> list[str]:
        """Return key fragments constrained by `equals` match rules.

        Args:
            equals: The equals constraint specified by the rule.
            text: The original command text to fall back on for sets.

        Returns:
            A list containing zero or one key fragments.

        """
        if equals is None:
            return []
        if isinstance(equals, str):
            return [f"equals|{equals}"]
        return [f"equals|{text}"]

    def _key_from_prefix(
        self,
        prefix: str | tuple[str, ...] | None,
        normalized_text: str,
    ) -> list[str]:
        """Return key fragments for `startswith` match rules.

        Args:
            prefix: The `startswith` constraint(s) to evaluate.
            normalized_text: The command text without the negation prefix.

        Returns:
            A list containing zero or one key fragments.

        """
        if prefix is None:
            return []
        matched = self._match_prefix(normalized_text, prefix)
        if matched is None:
            return []
        return [f"startswith|{matched}"]

    def _key_from_suffix(
        self,
        suffix: str | tuple[str, ...] | None,
        normalized_text: str,
    ) -> list[str]:
        """Return key fragments for `endswith` match rules.

        Args:
            suffix: The `endswith` constraint(s) to evaluate.
            normalized_text: The command text without the negation prefix.

        Returns:
            A list containing zero or one key fragments.

        """
        if suffix is None:
            return []
        matched = self._match_suffix(normalized_text, suffix)
        if matched is None:
            return []
        return [f"endswith|{matched}"]

    def _key_from_contains(
        self,
        contains: str | tuple[str, ...] | None,
        normalized_text: str,
    ) -> list[str]:
        """Return key fragments for `contains` match rules.

        Args:
            contains: The `contains` constraint(s) to evaluate.
            normalized_text: The command text without the negation prefix.

        Returns:
            A list containing zero or one key fragments.

        """
        if contains is None:
            return []
        matched = self._match_contains(normalized_text, contains)
        if matched is None:
            return []
        return [f"contains|{matched}"]

    def _key_from_regex(
        self,
        pattern: str | None,
        normalized_text: str,
        original_text: str,
    ) -> list[str]:
        """Return key fragments derived from regex match rules.

        Args:
            pattern: The regex pattern to match.
            normalized_text: The command text without the negation prefix.
            original_text: The command text including any negation.

        Returns:
            A list containing zero or one key fragments.

        """
        if pattern is None:
            return []

        match = search(pattern, normalized_text)
        match_source = normalized_text
        if match is None:
            match = search(pattern, original_text)
            match_source = original_text

        if match is None:
            return []

        regex_key = self._normalize_regex_key(pattern, match_source, match)
        return [f"re|{regex_key}"]

    @staticmethod
    def _match_prefix(value: str, prefix: str | tuple[str, ...]) -> str | None:
        if isinstance(prefix, tuple):
            matches = [candidate for candidate in prefix if value.startswith(candidate)]
            if matches:
                return max(matches, key=len)
            return None

        if value.startswith(prefix):
            return prefix

        return None

    @staticmethod
    def _match_suffix(value: str, suffix: str | tuple[str, ...]) -> str | None:
        if isinstance(suffix, tuple):
            matches = [candidate for candidate in suffix if value.endswith(candidate)]
            if matches:
                return max(matches, key=len)
            return None

        if value.endswith(suffix):
            return suffix

        return None

    @staticmethod
    def _match_contains(value: str, contains: str | tuple[str, ...]) -> str | None:
        if isinstance(contains, tuple):
            matches = [candidate for candidate in contains if candidate in value]
            if matches:
                return max(matches, key=len)
            return None

        if contains in value:
            return contains

        return None

    @staticmethod
    def _normalize_regex_key(pattern: str, value: str, match: Match[str]) -> str:
        """Normalize regex matches so equivalent commands hash the same."""
        result = match.group(0)

        if match.re.groups:
            groups = tuple(g or "" for g in match.groups())
            if any(groups):
                normalized_groups = tuple(group.strip() for group in groups)
                if any(normalized_groups):
                    return "|".join(normalized_groups)

        trimmed_pattern = pattern.rstrip("$")
        for suffix in (".*", ".+"):
            if trimmed_pattern.endswith(suffix):
                candidate_pattern = trimmed_pattern[: -len(suffix)]
                if not candidate_pattern:
                    break
                trimmed_match = search(candidate_pattern, value)
                if trimmed_match is not None:
                    candidate = trimmed_match.group(0).strip()
                    if candidate:
                        return candidate
                break

        return result.strip()

    @property
    def declaration_prefix(self) -> str:
        return ""

    @property
    def negation_prefix(self) -> str:
        return "no "

    @staticmethod
    def config_preprocessor(config_text: str) -> str:
        return config_text

    @staticmethod
    @abstractmethod
    def _instantiate_rules() -> HConfigDriverRules:
        pass
